import argparse
from pysam import VariantFile
from collections import defaultdict


def correct_positions_div(input_vcf):
    # reads input vcf, adjusts POS, END, TARGET values to account for previous events in the reference
    in_vcf = VariantFile(input_vcf)
    hd = in_vcf.header
    file_suffix = '_decoy_divrepeat_intervals.vcf'
    out_vcf = VariantFile(input_vcf[:-4] + file_suffix, 'w', header=hd)
    # store vcf records in dictionary keyed on start position
    vcf_recs = defaultdict(dict)
    for rec in in_vcf.fetch():
        vcf_recs[rec.chrom][rec.start] = rec

    # aggregate length of insertions up to a given point in the reference (to be added to the positions in a given vcf record)
    for chrom in vcf_recs.keys():
        total_ins_len = 0
        for pos in sorted(vcf_recs[chrom].keys()):
            evt = vcf_recs[chrom][pos]
            if evt.id == 'div_dDUP':
                evt.start += total_ins_len
                evt.info['TARGET'] += total_ins_len
                total_ins_len += evt.info['SVLEN']
        for rec in vcf_recs[chrom].values():
            # output separate records for each A and A* ("B") to serve as avoid intervals for further simulation
            # ** assuming vcf will only contain SVs of type div_dDUP
            rec_A = rec.copy()
            rec_B = rec.copy()
            rec_A.id = 'div_dDUP_A'
            rec_A.alts = ('div_dDUP_A',)
            rec_A.info['SVTYPE'] = 'div_dDUP_A'
            rec_A.info['TARGET'] = -1
            # setting the above unsets rec.stop, so need to reset it to the original
            rec_A.stop = rec.stop
            # if the event is flipped, then need to shift the _A interval by one interval length
            # to compensate for the insertion at the target locus, that will in this case happen before the source
            if rec.info['TARGET'] < rec.start:
                rec_A.start += rec.info['SVLEN']
            rec_B.id = 'div_dDUP_B'
            rec_B.alts = ('div_dDUP_B',)
            rec_B.start = rec_B.info['TARGET']
            rec_B.stop = rec_B.info['TARGET'] + rec.info['SVLEN']
            rec_B.info['SVTYPE'] = 'div_dDUP_B'
            rec_B.info['TARGET'] = -1
            out_vcf.write(rec_A)
            out_vcf.write(rec_B)

    out_vcf.close()
    in_vcf.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Utility script for multi-staged simulation procedure of divergent repeats')
    parser.add_argument('--div_dDUP_vcf', help='Input vcf from R1 generation step')
    args = parser.parse_args()
    correct_positions_div(args.div_dDUP_vcf)



from enum import Enum

MAX_BUFFER_SIZE: int = 1000000  # max number of bases that can be read at one time to export to fasta file


class Variant_Type(Enum):
    INS = "INS"
    DEL = "DEL"
    INV = "INV"
    DUP = "DUP"
    SNP = "SNP"
    TRA = "TRA"
    dupINVdup = "dupINVdup"
    delINVdel = "delINVdel"
    delINVdup = "delINVdup"
    dupINVdel = "dupINVdel"
    delINV = "delINV"
    INVdel = "INVdel"
    dDUP_iDEL = "dDUP_iDEL"
    INS_iDEL = "INS_iDEL"
    dup_INV = "dup_INV"
    INV_dup = "INV_dup"
    INVdup = "INVdup"
    dDUP = "dDUP"
    Custom = "Custom"
    INV_dDUP = "INV_dDUP"
    div_dDUP = "div_dDUP"
    DIVERGENCE = "DIVERGENCE"


# list of dispersion-based events (listing explicitly to prevent having to parse from symbols)
DISPERSION_TYPES = [Variant_Type.dDUP, Variant_Type.INV_dDUP,
                    Variant_Type.TRA, Variant_Type.div_dDUP,
                    Variant_Type.dDUP_iDEL, Variant_Type.INS_iDEL]


class Operations(Enum):
    INS = "INS"
    DUP = "DUP"
    INV = "INV"
    DEL = "DEL"
    TRA = "TRA"
    INVDUP = "INVDUP"
    INVTRA = "INVTRA"
    IDENTITY = "IDENTITY"
    UNDEFINED = "UNDEFINED"
    DIV = "DIV"


# list of operations corresponding to events that should have bed record with order > 0
NONZERO_ORDER_OPERATIONS = [Operations.TRA.value, Operations.INS.value, Operations.DUP.value, Operations.INVDUP.value,
                            Operations.INVTRA.value, Operations.DIV.value]


class Zygosity(Enum):
    UNDEFINED = -1
    HOMOZYGOUS = 1
    HETEROZYGOUS = 0


class Symbols(Enum):
    DIS = "_"  # dispersion event
    DUP = "'"  # attached to symbols that are not the original one from source sequence
    DIV = "*"  # divergent interval, attached to symbols that vary from the original by low-probability base error
    REQUIRED_SOURCE = "A"  # event symbol of the required source/main event all SVs must have


SV_KEY = {Variant_Type.INS: [(), ("A")],
          Variant_Type.SNP: [("A",), ("A*",)],
          Variant_Type.DEL: [("A",), ()],
          Variant_Type.INV: [("A",), ("a",)],
          Variant_Type.DUP: [("A",), ("A", "A'")],
          Variant_Type.TRA: [("A", "_"), ("_", "A'")],
          Variant_Type.dupINVdup: [("A", "B", "C"), ("A", "c'", "b", "a'", "C")],
          Variant_Type.delINVdel: [("A", "B", "C"), ("b",)],
          Variant_Type.delINVdup: [("A", "B", "C"), ("c'", "b", "C")],
          Variant_Type.dupINVdel: [("A", "B", "C"), ("A", "b", "a'")],
          Variant_Type.delINV: [("A", "B"), ("b",)],
          Variant_Type.INVdel: [("A", "B"), ("a",)],
          Variant_Type.dDUP_iDEL: [("A", "_", "B"), ("A", "_", "A'")],
          Variant_Type.INS_iDEL: [("A", "_", "B"), ("_", "A'")],
          Variant_Type.INVdup: [("A",), ("a", "a'")],
          Variant_Type.dup_INV: [("A", "B"), ("A", "b", "a'")],
          Variant_Type.INV_dup: [("A", "B"), ("b'", "a", "B")],
          Variant_Type.dDUP: [("A", "_"), ("A", "_", "A'")],
          Variant_Type.INV_dDUP: [("A", "_"), ("A", "_", "a'")],
          Variant_Type.div_dDUP: [("A", "_"), ("A", "_", "A*")],
          Variant_Type.DIVERGENCE: [("A",), ("A*",)]}

DEFAULT_CONFIG = {"sim_settings": {"max_tries": 100,
                                   "fail_if_placement_issues": False,
                                   "generate_log_file": False,
                                   "filter_small_chr": True,
                                   "prioritize_top": False},
                  "SVs": {}}



import os
import random
import itertools
import copy  # <- utils for making copies of python objects
import argparse
import numpy as np
from insilicosv.constants import *
from collections import defaultdict
from pysam import VariantFile


class NestedDict(defaultdict):
    def __call__(self):
        return NestedDict(self.default_factory)


def get_original_base_pos(genome_vcf, query_loc, query_chrom):
    # calculates the input reference position of the base that corresponds
    # to the one at position query_loc in the output reference (i.e., adjusts the
    # input ref position by the cumulative change in ref length caused by
    # the variants from [0, query_loc)
    # **for now: just admits DEL, DUP, INV
    vcf = VariantFile(genome_vcf)
    # need to make sure vcf recs sorted by position and separated by chrom
    vcf_recs = defaultdict(list)
    for rec in vcf.fetch():
        vcf_recs[rec.chrom].append((rec.id, rec.start, rec.stop))
    for chrom in vcf_recs.keys():
        vcf_recs[chrom].sort(key=lambda x: x[1])
    # total the cumulative change in ref length from all the SVs *ending* before p
    p = query_loc
    i = 0
    while i < len(vcf_recs[query_chrom]) and vcf_recs[query_chrom][i][1] <= p:
        rec_id, rec_start, rec_stop = vcf_recs[query_chrom][i]
        rec_len = rec_stop - rec_start
        if rec_id == 'DEL' and rec_start <= p:
            p += rec_len
        if rec_id == 'DUP' and rec_stop <= p:
            p -= rec_len
        i += 1
    # check if p is internal to an INV; edge case not covered by the above loop
    surrounding_inv = next((rec for rec in vcf_recs[query_chrom] if rec[1] <= p < rec[2] and rec[0] == 'INV'), None)
    if surrounding_inv is not None:
        sv_id, sv_start, sv_end = surrounding_inv
        p += (sv_end - sv_start) - ((p - sv_start) * 2 + 1)
    return p


def is_overlapping(event_ranges, addition, called_from_helper=False, strictly_partial=False):
    # addition: tuple (start, end)
    # event_ranges: list containing tuples
    # checks if addition overlaps with any of the events already stored
    # 'strictly_partial' a toggle to return True if the interval overlaps a stored event, without being equal to it
    for event in event_ranges:
        if event[1] > addition[0] and event[0] < addition[1]:
            if called_from_helper:
                raise Exception("Overlap between {} and {}".format(event[0:2], addition[0:2]))
            else:
                return True if not strictly_partial else event != addition

    return False


def fail_if_any_overlapping(arr):
    # will raise Exception if any overlap between intervals is found
    # arr: list of tuples
    for x, ele in enumerate(arr):
        is_overlapping(arr[:x], ele, called_from_helper=True)


def validate_symbols(source, target):
    '''
    Ensures that source has unique symbols other than dispersion events - specifically used for source sequence
    Raises error if number of dispersions between source and target is not the same

    source: tuple, source sequence
    target: tuple, target sequence
    '''
    present = dict()
    for symbol in source:
        if symbol in present:
            raise Exception("Source transformation {} does not have unique symbols!".format(source))
        elif symbol != Symbols.DIS.value:   # exclude dispersion events because they will always appear the same for user inputs
            present[symbol] = True

        if Symbols.DUP.value in symbol:
            raise Exception("Duplication marking (') not allowed in source sequence {}".format(source))
        if any(c.islower() for c in symbol):
            raise Exception("Only original symbols may appear in source sequence, and they must also be uppercase: {}".format(source))
    
    assert sum([1 for symbol in source if symbol.startswith(Symbols.DIS.value)]) == sum([1 for symbol in target if symbol.startswith(Symbols.DIS.value)]), "Number of dispersion events must be equal between source and target sequence"


def remove_file(file):
    if os.path.exists(file):
        os.remove(file)


def reset_file(filename):
    with open(filename, "w") as f_reset:
        f_reset.truncate()


def generate_seq(length):
    base_map = {1: "A", 2: "T", 3: "G", 4: "C"}
    return ''.join([base_map[random.randint(1, 4)] for _ in range(length)])


def percent_N(seq):
    return 0 if len(seq) == 0 else seq.count('N') / len(seq)


def complement(seq):
    output = ""
    base_complements = {"A": "T", "T": "A", "G": "C", "C": "G", "N": "N"}
    for base in seq.upper():
        if base in base_complements:
            output += base_complements[base]
        else:
            output += base

    return output


def divergence(seq, divergence_prob=None):
    # apply random base flips to input sequence
    # probability of changing the base, chosen from U(0.5,1.0) or given from user
    p = random.uniform(0.5, 1.0) if divergence_prob is None else float(divergence_prob)
    return ''.join([b if random.random() > p else random.choice(list({"A", "C", "T", "G"} - {b})) for b in seq.upper()])


def get_sv_config_identifier(sv_config):
    # helper function to yield a uniquely identifying string for a given SV config entry
    # --> usage: needing to index into overlap events dicts to access info specific to a set of SVs simulated with a given set of params
    min_str = str(sv_config['min_length']) if isinstance(sv_config['min_length'], int) else '_'.join(map(str, sv_config['min_length']))
    max_str = str(sv_config['max_length']) if isinstance(sv_config['max_length'], int) else '_'.join(map(str, sv_config['max_length']))
    return sv_config['type'].value + '_' + min_str + '_' + max_str


class OverlapEvents:
    # container object for the optionally-provided genome context elements that will be used in SV placement
    def __init__(self, config, allow_chroms=None):
        # elements will be stored in a dict keyed on element name
        self.overlap_events_dict = defaultdict(list)
        self.allow_chroms = allow_chroms
        self.allow_types = None if 'allow_types' not in config['overlap_events'].keys() else config['overlap_events']['allow_types']
        if isinstance(self.allow_types, str):
            self.allow_types = [self.allow_types]
        self.svtype_overlap_counts = defaultdict(NestedDict(int))
        self.svtype_partial_overlap_counts = defaultdict(NestedDict(int))
        self.svtype_alu_mediated_counts = defaultdict(NestedDict(int))
        self.get_num_overlap_counts(config)

        if type(config['overlap_events']['bed']) is list:
            for bed_path in config['overlap_events']['bed']:
                self.parse_bed_file(bed_path, allow_chroms=self.allow_chroms, allow_types=self.allow_types)
        else:
            self.parse_bed_file(config['overlap_events']['bed'], allow_chroms=self.allow_chroms, allow_types=self.allow_types)

        # optional list of Alu pairs to be used as flanking Alus for DELs/DUPs (to simulate Alu-mediated CNVs)
        # --> dict of form {SV_identifier: [(chrom_1, a_1, b_1), ..., (chrom_n, a_n, b_n)]}
        self.alu_pairs = self.populate_alu_pairs(config['SVs'])

    def get_num_overlap_counts(self, config):
        # populate nested dict of the form {sv type: {element type: num_overlap}}
        for sv in config['SVs']:
            sv_config_key = get_sv_config_identifier(sv)
            # building separate counts dictionaries for complete overlaps, partial overlaps, and alu-mediated intervals
            for overlap_count, count_dict in [('num_overlap', self.svtype_overlap_counts), ('num_partial_overlap', self.svtype_partial_overlap_counts)]:
                if overlap_count in sv.keys():
                    if isinstance(sv[overlap_count], int) and sv[overlap_count] > 0:
                        # set overlap type label based on specified allow_types
                        if self.allow_types is not None and len(self.allow_types) == 1:
                            count_dict[sv_config_key][self.allow_types[0]] = sv[overlap_count]
                        else:
                            count_dict[sv_config_key]['ALL'] = sv[overlap_count]
                    else:
                        # define correspondence with the element types as they were given in the config file
                        for i in range(len(sv[overlap_count])):
                            if sv[overlap_count][i] > 0:  # <- only counting entries in list > 0
                                count_dict[sv_config_key][self.allow_types[i]] = sv[overlap_count][i]
            if 'num_alu_mediated' in sv.keys():
                self.svtype_alu_mediated_counts[sv_config_key] = sv['num_alu_mediated']

    def parse_bed_file(self, bed_fname, allow_chroms=None, allow_types=None):
        """
        reads bed file (intended use: processing repeatmasker elements to be considered for randomized
        event overlap) and adds (chr, start, end) tuples representing the intervals to a dict keyed on elt type
        - allow_chroms: optional list of allowed chromosomes (filtering out all entries with chrom not in list)
        - allow_types: optional list of allowed types to be included in the simulation
        """
        with open(bed_fname, 'r') as bed_file:
            for line in bed_file:
                if line.startswith('#') or line.isspace():
                    continue
                fields = line.strip().split()
                assert len(fields) >= 3, "Unexpected number of fields in BED: %s" % line
                chr_name, start, end, elt_type = fields[:4]
                # need to allow for allow_types to include prefixes of element names (always including Alu as an allowed
                # prefix in case any of the SVs will need to be placed at Alu-mediated intervals)
                if (allow_chroms and chr_name not in allow_chroms) or (allow_types and not any([elt_name in elt_type for elt_name in allow_types + ['Alu']])):
                    continue
                self.overlap_events_dict[elt_type].append((chr_name, int(start), int(end)))
        # shuffle the lists of each element type for random selection
        for k in self.overlap_events_dict.keys():
            random.shuffle(self.overlap_events_dict[k])

    def get_single_element_interval(self, sv_config_id, sv_config, partial_overlap):
        if len(self.overlap_events_dict) == 0:
            return None, None, None
        counts_dict = self.svtype_partial_overlap_counts if partial_overlap else self.svtype_overlap_counts
        elt_type = list(counts_dict[sv_config_id].keys())[0]
        # if this elt_type doesn't have any matching elts in the element dict, remove it and move on to the next elt
        while elt_type != 'ALL' and not any([elt_type in ovlp_type for ovlp_type in self.overlap_events_dict.keys()]):
            del counts_dict[sv_config_id][elt_type]
            if len(counts_dict[sv_config_id]) == 0:
                del counts_dict[sv_config_id]
                break
            elt_type = list(counts_dict[sv_config_id].keys())[0]
        # if num_overlaps was given as a single number rather than an element-specific list, draw a random element for overlap
        repeat_elt, retrieved_type = self.__getitem__(sv_config_id=sv_config_id,
                                                      minsize=sv_config['length_ranges'][0][0],
                                                      maxsize=sv_config['length_ranges'][0][1],
                                                      elt_type=(None if elt_type == 'ALL' else elt_type),
                                                      partial_overlap=partial_overlap)
        return repeat_elt, retrieved_type, elt_type

    def populate_alu_pairs(self, svs_config):
        if not any('num_alu_mediated' in d.keys() for d in svs_config):
            return None
        # construct dict of form {chrom: [(start, end)]} Alu interval tuples
        alu_intervals = defaultdict(list)
        for elt_type, elt_list in self.overlap_events_dict.items():
            if 'Alu' in elt_type:
                for chrom, start, end in elt_list:
                    alu_intervals[chrom].append((int(start), int(end)))
        for chrom in alu_intervals.keys():
            random.shuffle(alu_intervals[chrom])
        alu_pairs_dict = {}
        # construct the lists of viable flanking pairs for each SV config with a nonzero num_alu_mediated specified
        for sv_config in svs_config:
            sv_config_id = get_sv_config_identifier(sv_config)
            if 'num_alu_mediated' in sv_config.keys():
                alu_intervals_copy = copy.deepcopy(alu_intervals)
                alu_pairs = []
                sv_min, sv_max = sv_config['min_length'], sv_config['max_length']
                for _ in range(self.svtype_alu_mediated_counts[sv_config_id]):
                    viable_matches = []
                    while len(viable_matches) == 0:
                        if len(alu_intervals_copy.keys()) == 0 or not any(len(v) > 1 for v in alu_intervals_copy.values()):
                            break  # <-- no Alus remaining to choose from
                        # choose a random starting (left) Alu
                        rand_chrom = random.choice([chrom for chrom, alus in alu_intervals_copy.items() if len(alus) > 1])
                        alu1 = alu_intervals_copy[rand_chrom].pop()
                        # collect all the alus on the same chromosome that are within an appropriate dist
                        viable_matches = [ivl for ivl in alu_intervals_copy[rand_chrom] if sv_min <= np.abs(self.midpoint(*ivl) - self.midpoint(*alu1)) <= sv_max]
                    if len(viable_matches) > 0:
                        alu2 = random.choice(viable_matches)
                        alu_intervals_copy[rand_chrom].remove(alu2)
                        if len(alu_intervals_copy[rand_chrom]) == 0:
                            del alu_intervals_copy[rand_chrom]
                        alu_pairs.append((rand_chrom, alu1, alu2) if alu1[0] < alu2[0] else (rand_chrom, alu2, alu1))
                        # delete the used Alus from the alu_intervals dict and from the general overlap_events_dict
                        for alu in [alu1, alu2]:
                            alu_intervals[rand_chrom].remove(alu)
                            self.remove_alu_from_overlap_dict(rand_chrom, *alu)
                if len(alu_pairs) > 0:
                    random.shuffle(alu_pairs)
                    alu_pairs_dict[sv_config_id] = alu_pairs
        return alu_pairs_dict

    def get_alu_mediated_interval(self, sv_config_id):
        if len(self.alu_pairs[sv_config_id]) == 0:
            del self.svtype_alu_mediated_counts[sv_config_id]
            return None, None
        ivl_chrom, left_alu, right_alu = self.alu_pairs[sv_config_id].pop()
        ivl_start, ivl_end = (left_alu[1] + left_alu[0]) // 2, (right_alu[1] + right_alu[0]) // 2
        self.svtype_alu_mediated_counts[sv_config_id] -= 1
        if self.svtype_alu_mediated_counts[sv_config_id] == 0:
            del self.svtype_alu_mediated_counts[sv_config_id]
        return (ivl_chrom, ivl_start, ivl_end), 'ALU_MEDIATED'  # <- ALU_MEDIATED to be used as retrieved_type

    def remove_alu_from_overlap_dict(self, chrom, start, end):
        # helper method to remove an Alu element from the overlap_events_dict if it's been chosen for an alu-mediated interval
        for k, v in self.overlap_events_dict.items():
            if k.startswith('Alu') and (chrom, start, end) in v:
                self.overlap_events_dict[k].remove((chrom, start, end))
        self.overlap_events_dict = {k: v for k, v in self.overlap_events_dict.items() if len(v) > 0}

    @staticmethod
    def midpoint(start, end):
        return (start + end) // 2

    @staticmethod
    def get_intrvl_len(chr, st, end):
        return int(end) - int(st)

    def elt_type_is_allowed(self, elt_type):
        # helper to check if an element type is allowed by the allow_types list
        return True if self.allow_types is None else any(allow_type.startswith(elt_type) for allow_type in self.allow_types)

    @staticmethod
    def get_partially_overlapping_interval(elt_chrom, elt_start, elt_stop, sv_min, sv_max):
        # choose sv size, draw a random position in element interval, randomly set that to the start or end of SV
        def draw_from_unif(a, b):
            return a if a == b else np.random.randint(a, b)
        sv_size = draw_from_unif(sv_min, sv_max)
        interval_left = draw_from_unif(max(elt_start - sv_size + 1, 0), max(elt_stop - sv_size, 0)) if random.randint(0, 1)\
            else draw_from_unif(elt_start, max(elt_stop - sv_size - 1, elt_start))  # <- two possible distributions for the left bound of the SV
        return elt_chrom, interval_left, interval_left + sv_size

    def decrement_counts(self, sv_config_id, input_elt_type, partial_overlap):
        # decrement the relevant counts dict based on element extracted in __getitem__()
        counts_dict = self.svtype_partial_overlap_counts if partial_overlap else self.svtype_overlap_counts
        counts_dict[sv_config_id][input_elt_type] -= 1
        if counts_dict[sv_config_id][input_elt_type] == 0:
            del counts_dict[sv_config_id][input_elt_type]
            if len(counts_dict[sv_config_id].keys()) == 0:
                del counts_dict[sv_config_id]

    def __getitem__(self, sv_config_id, minsize, maxsize, elt_type=None, partial_overlap=False):
        input_elt_type = 'ALL' if elt_type is None else elt_type
        if elt_type in self.overlap_events_dict.keys():  # <- elt_type given, and elements of matching type are in the dict
            rand_elt = next((elt for elt in self.overlap_events_dict[elt_type] if minsize <= self.get_intrvl_len(*elt) <= maxsize), None) \
                        if not partial_overlap else next(elt for elt in self.overlap_events_dict[elt_type])
        elif len(self.overlap_events_dict.keys()) > 0:  # <- element dict is non-empty; elt_type not given, or is given as a prefix
            if elt_type is None:
                elt_type_mapping = {elt: elt_type for (elt_type, elt_list) in self.overlap_events_dict.items() for elt in elt_list}
                # elt_type-agnostic branch: draw random element from combined list of elements across all types
                # and if allow_types is specified, only grab agreeing elements
                rand_elt = next((elt for elt in itertools.chain.from_iterable(self.overlap_events_dict.values()) if
                                 (minsize <= self.get_intrvl_len(*elt) <= maxsize or partial_overlap) and
                                 (self.elt_type_is_allowed(elt_type_mapping[elt]))), None)
            else:
                # elt_type prefix branch: choose an element whose name begins with elt_type
                prefix_matches = [element_type for element_type in self.overlap_events_dict.keys() if element_type.startswith(elt_type)]
                if len(prefix_matches) > 0:
                    random.shuffle(prefix_matches)
                    elt_type = prefix_matches.pop()
                    while len(prefix_matches) > 0 and len([elt for elt in self.overlap_events_dict[elt_type] if minsize <= self.get_intrvl_len(*elt) <= maxsize]) == 0:
                        elt_type = prefix_matches.pop()
                    rand_elt = next((elt for elt in self.overlap_events_dict[elt_type] if minsize <= self.get_intrvl_len(*elt) <= maxsize), None) \
                        if not partial_overlap else next(elt for elt in self.overlap_events_dict[elt_type])
                else:
                    rand_elt = None
        else:  # <- elt_type is NOT given and there AREN'T elements in the dict
            rand_elt = None
        if rand_elt is not None:
            if elt_type is None:
                elt_type = elt_type_mapping[rand_elt]
            self.overlap_events_dict[elt_type].remove(rand_elt)
            if len(self.overlap_events_dict[elt_type]) == 0:
                del self.overlap_events_dict[elt_type]
            self.decrement_counts(sv_config_id, input_elt_type, partial_overlap)
        elif elt_type in self.overlap_events_dict.keys():  # <- case in which elements of valid type but not size were available
            self.decrement_counts(sv_config_id, input_elt_type, partial_overlap)
        if partial_overlap and rand_elt is not None:
            rand_elt = self.get_partially_overlapping_interval(*rand_elt, minsize, maxsize)
        return rand_elt, elt_type


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Special use case for utils.py: Commandline access to get_original_base_pos()')
    parser.add_argument('--genome_vcf', type=str, help='Path to vcf describing simulated SVs populating the synthetic genome')
    parser.add_argument('--query_loc', type=int, help='Query genome locus (given with respect to synthetic reference)')
    parser.add_argument('--query_chrom', type=int, help='Corresponding chromosome of the query locus')
    args = parser.parse_args()

    print('Input reference position corresponding to mutated reference position (%s) %d: %d' %
          (args.query_chrom, args.query_loc, get_original_base_pos(args.genome_vcf, args.query_loc, args.query_chrom)))




from insilicosv.simulate import SV_Simulator
from insilicosv.processing import FormatterIO
from test_simulate import TestObject
from pysam import VariantFile, FastaFile
from collections import defaultdict, Counter
from insilicosv.utils import NestedDict
import unittest
import sys
import os
from insilicosv import utils
from insilicosv import constants


class TestProcObject(TestObject):
    def __init__(self, ref, par, hap1, hap2, bed, vcf):
        self.vcf = vcf
        super().__init__(ref, par, hap1, hap2, bed)

    def extract_bed_records(self):
        # parse bed record into dict for easy comparison
        # --> example split bed record: ['chr19', '0', '3', 'chr19', '0', '3', 'DEL', '3', '1/1', 'DEL', '1', '0']
        bed_records = []
        with open(self.bed) as f:
            for line in f:
                ln = line.split()
                bed_record = {'source_chr': ln[0], 'source_s': ln[1], 'source_e': ln[2],
                              'target_chr': ln[3], 'target_s': ln[4], 'target_e': ln[5],
                              'ev_type': ln[6], 'len': ln[7], 'zyg': ln[8], 'parent_type': ln[9],
                              'nth_sv': ln[10], 'order': ln[11]}
                bed_records.append(bed_record)
        return bed_records

    def extract_vcf_records(self):
        vcf_records = []
        vcf = VariantFile(self.vcf)
        for rec in vcf.fetch():
            ln = str(rec).split()
            # separately parse info field of the form: 'END=45590417;SVTYPE=dDUP;SVLEN=539;TARGET=45581738'
            info = {field.split('=')[0]: field.split('=')[1] for field in ln[7].split(';')}
            vcf_record = {'CHROM': ln[0], 'POS': ln[1], 'ID': ln[2], 'REF': ln[3], 'ALT': ln[4], 'QUAL': ln[5],
                          'FILTER': ln[6], 'INFO': info, 'FORMAT': ln[8], 'SAMPLE': ln[9]}
            vcf_records.append(vcf_record)
        return vcf_records


class TestProcessing(unittest.TestCase):
    def setUp(self):
        # runs before every test
        self.ref_file = "test/inputs/test.fa"
        self.par = "test/inputs/par.yaml"

        self.hap1 = "test/inputs/test1.fa"
        self.hap2 = "test/inputs/test2.fa"
        self.bed = "test/inputs/out.bed"
        self.vcf = "test/inputs/out.vcf"
        self.ins_fasta = "test/inputs/ins_fasta.fa"
        self.test_overlap_bed = "test/inputs/example_overlap_events.bed"
        self.test_overlap_bed_2 = "test/inputs/example_overlap_events_2.bed"
        # test_overlap_bed_3: events with differing chromosome
        self.test_overlap_bed_3 = "test/inputs/example_overlap_events_3.bed"
        self.test_overlap_bed_4 = "test/inputs/example_overlap_events_4.bed"
        self.test_overlap_bed_11 = "test/inputs/example_overlap_events_11.bed"

        self.test_objects_simple_events = {'DEL': TestProcObject([self.ref_file, {"chr19": "CTG"}],
                                                                 [self.par, {"sim_settings": {"max_tries": 50, "prioritize_top": True},
                                                                             "SVs": [{"type": "DEL", "number": 1,
                                                                                      "max_length": 3,
                                                                                      "min_length": 3}]}],
                                                                 self.hap1, self.hap2, self.bed, self.vcf),
                                           'DUP': TestProcObject([self.ref_file, {"chr19": "CTG"}],
                                                                 [self.par, {"sim_settings": {"max_tries": 50, "prioritize_top": True},
                                                                             "SVs": [{"type": "DUP", "number": 1,
                                                                                      "max_length": 3,
                                                                                      "min_length": 3}]}],
                                                                 self.hap1, self.hap2, self.bed, self.vcf),
                                           'INV': TestProcObject([self.ref_file, {"chr19": "CTG"}],
                                                                 [self.par, {"sim_settings": {"max_tries": 50, "prioritize_top": True},
                                                                             "SVs": [{"type": "INV", "number": 1,
                                                                                      "max_length": 3,
                                                                                      "min_length": 3}]}],
                                                                 self.hap1, self.hap2, self.bed, self.vcf),
                                           'INS': TestProcObject([self.ref_file, {"chr19": "C"}],
                                                                 [self.par, {"sim_settings": {"max_tries": 50, "prioritize_top": True},
                                                                             "SVs": [{"type": "INS", "number": 1,
                                                                                      "max_length": 3,
                                                                                      "min_length": 3}]}],
                                                                 self.hap1, self.hap2, self.bed, self.vcf)}
        self.test_objects_flanked_inversions = {'dupINVdup': TestProcObject([self.ref_file, {"chr19": "ACTGTC"}],
                                                                            [self.par,
                                                                             {"sim_settings": {"prioritize_top": True},
                                                                              "SVs": [{"type": "dupINVdup", "number": 1,
                                                                                       "max_length": [2, 2, 2],
                                                                                       "min_length": [2, 2, 2]}]}],
                                                                            self.hap1, self.hap2, self.bed, self.vcf),
                                                'delINVdel': TestProcObject([self.ref_file, {"chr19": "ACTGTC"}],
                                                                            [self.par,
                                                                             {"sim_settings": {"prioritize_top": True},
                                                                              "SVs": [{"type": "delINVdel", "number": 1,
                                                                                       "max_length": [2, 2, 2],
                                                                                       "min_length": [2, 2, 2]}]}],
                                                                            self.hap1, self.hap2, self.bed, self.vcf),
                                                'dupINVdel': TestProcObject([self.ref_file, {"chr19": "ACTGTC"}],
                                                                            [self.par,
                                                                             {"sim_settings": {"prioritize_top": True},
                                                                              "SVs": [{"type": "dupINVdel", "number": 1,
                                                                                       "max_length": [2, 2, 2],
                                                                                       "min_length": [2, 2, 2]}]}],
                                                                            self.hap1, self.hap2, self.bed, self.vcf),
                                                'delINVdup': TestProcObject([self.ref_file, {"chr19": "ACTGTC"}],
                                                                            [self.par,
                                                                             {"sim_settings": {"prioritize_top": True},
                                                                              "SVs": [{"type": "delINVdup", "number": 1,
                                                                                       "max_length": [2, 2, 2],
                                                                                       "min_length": [2, 2, 2]}]}],
                                                                            self.hap1, self.hap2, self.bed, self.vcf)}
        self.test_objects_dispersions = {'dDUP': TestProcObject([self.ref_file, {"chr19": "ACTGTC"}],
                                                                [self.par,
                                                                 {"sim_settings": {"prioritize_top": True},
                                                                  "SVs": [{"type": "dDUP", "number": 1,
                                                                           "max_length": [3, 3],
                                                                           "min_length": [3, 3]}]}],
                                                                self.hap1, self.hap2, self.bed, self.vcf),
                                         'INV_dDUP': TestProcObject([self.ref_file, {"chr19": "ACTGTC"}],
                                                                    [self.par,
                                                                     {"sim_settings": {"prioritize_top": True},
                                                                      "SVs": [{"type": "INV_dDUP", "number": 1,
                                                                               "max_length": [3, 3],
                                                                               "min_length": [3, 3]}]}],
                                                                    self.hap1, self.hap2, self.bed, self.vcf),
                                         'TRA': TestProcObject([self.ref_file, {"chr19": "ACTGTC"}],
                                                               [self.par,
                                                                {"sim_settings": {"prioritize_top": True},
                                                                 "SVs": [{"type": "TRA", "number": 1,
                                                                          "max_length": [3, 3],
                                                                          "min_length": [3, 3]}]}],
                                                               self.hap1, self.hap2, self.bed, self.vcf)}
        self.test_objects_del_inv = {'delINV': TestProcObject([self.ref_file, {"chr19": "ACTGTC"}],
                                                              [self.par,
                                                               {"sim_settings": {"prioritize_top": True},
                                                                "SVs": [{"type": "delINV", "number": 1,
                                                                         "max_length": [3, 3],
                                                                         "min_length": [3, 3]}]}],
                                                              self.hap1, self.hap2, self.bed, self.vcf),
                                     'INVdel': TestProcObject([self.ref_file, {"chr19": "ACTGTC"}],
                                                              [self.par,
                                                               {"sim_settings": {"prioritize_top": True},
                                                                "SVs": [{"type": "INVdel", "number": 1,
                                                                         "max_length": [3, 3],
                                                                         "min_length": [3, 3]}]}],
                                                              self.hap1, self.hap2, self.bed, self.vcf)}
        self.test_objects_idel = {'dDUP_iDEL': TestProcObject([self.ref_file, {"chr19": "ACTGTCAG"}],
                                                              [self.par,
                                                               {"sim_settings": {"prioritize_top": True},
                                                                "SVs": [{"type": "dDUP_iDEL", "number": 1,
                                                                         "max_length": [3, 3, 2],
                                                                         "min_length": [3, 3, 2]}]}],
                                                              self.hap1, self.hap2, self.bed, self.vcf),
                                  'INS_iDEL': TestProcObject([self.ref_file, {"chr19": "ACTGTCAG"}],
                                                             [self.par,
                                                              {"sim_settings": {"prioritize_top": True},
                                                               "SVs": [{"type": "INS_iDEL", "number": 1,
                                                                        "max_length": [3, 3, 2],
                                                                        "min_length": [3, 3, 2]}]}],
                                                             self.hap1, self.hap2, self.bed, self.vcf)}
        self.test_objects_dup_inv = {'dup_INV': TestProcObject([self.ref_file, {"chr19": "ACTGTCAG"}],
                                                               [self.par,
                                                                {"sim_settings": {"prioritize_top": True},
                                                                 "SVs": [{"type": "dup_INV", "number": 1,
                                                                          "max_length": [4, 4],
                                                                          "min_length": [4, 4]}]}],
                                                               self.hap1, self.hap2, self.bed, self.vcf),
                                     'INV_dup': TestProcObject([self.ref_file, {"chr19": "ACTGTCAG"}],
                                                               [self.par,
                                                                {"sim_settings": {"prioritize_top": True},
                                                                 "SVs": [{"type": "INV_dup", "number": 1,
                                                                          "max_length": [4, 4],
                                                                          "min_length": [4, 4]}]}],
                                                               self.hap1, self.hap2, self.bed, self.vcf)}
        self.test_objects_INVdup = {'INVdup': TestProcObject([self.ref_file, {"chr19": "ACTG"}],
                                                             [self.par,
                                                              {"sim_settings": {"prioritize_top": True},
                                                               "SVs": [{"type": "INVdup", "number": 1,
                                                                        "max_length": 4,
                                                                        "min_length": 4}]}],
                                                             self.hap1, self.hap2, self.bed, self.vcf)}
        self.test_objects_multievent = {'INVdup': TestProcObject([self.ref_file, {"chr19": "ACTGCTAATGCGTTCACTGCTAATGCGTTC"}],
                                                                 [self.par,
                                                                  {"sim_settings": {"max_tries": 200, "prioritize_top": True},
                                                                   "SVs": [{"type": "INVdup", "number": 3,
                                                                            "max_length": 4,
                                                                            "min_length": 2}]}],
                                                                 self.hap1, self.hap2, self.bed, self.vcf)}
        self.test_objects_overlap_simple = {'overlap1': TestProcObject([self.ref_file, {"chr21": "CTCCGTCGTACTAAGTCGTACTCCGTCGTACTAAGTCGTA"}],
                                                                       [self.par, {"sim_settings": {"prioritize_top": True,
                                                                                                    "fail_if_placement_issues": True},
                                                                                   "overlap_events": {
                                                                                   "bed": [self.test_overlap_bed, self.test_overlap_bed_2],
                                                                                   "allow_types": ["L1HS", "ALR/Alpha"]},
                                                                                   "SVs": [{"type": "DEL", "number": 5,
                                                                                            "min_length": 1, "max_length": 5,
                                                                                            "num_overlap": [2, 1]}]}],
                                                                       self.hap1, self.hap2, self.bed, self.vcf),
                                            'overlap2': TestProcObject([self.ref_file, {"chr21": "CTCCGTCGTACTAAGTCGTACTCCGTCGTACTAAGTCGTA"}],
                                                                       [self.par, {"sim_settings": {"prioritize_top": True},
                                                                                   "overlap_events": {
                                                                                   "bed": [self.test_overlap_bed, self.test_overlap_bed_2],
                                                                                   "allow_types": ["L1HS", "ALR/Alpha"]},
                                                                                   "SVs": [{"type": "DEL", "number": 4,
                                                                                            "min_length": 1, "max_length": 5,
                                                                                            "num_overlap": [3, 1]}]}],
                                                                       self.hap1, self.hap2, self.bed, self.vcf),
                                            'overlap3': TestProcObject([self.ref_file, {"chr21": "CTCCGTCGTACTAAGTCGTACTCCGTCGTACTAAGTCGTA"}],
                                                                       [self.par, {"sim_settings": {"prioritize_top": True,
                                                                                                    "fail_if_placement_issues": True},
                                                                                   "overlap_events": {
                                                                                   "bed": [self.test_overlap_bed, self.test_overlap_bed_2],
                                                                                   "allow_types": ["L1", "ALR"]},
                                                                                   "SVs": [{"type": "DEL", "number": 5,
                                                                                            "min_length": 1, "max_length": 5,
                                                                                            "num_overlap": [3, 2]}]}],
                                                                       self.hap1, self.hap2, self.bed, self.vcf),
                                            'overlap4': TestProcObject([self.ref_file, {"chr21": "CTCCGTCGTACTAAGTCGTACTCCGTCGTACTAAGTCGTA"}],
                                                                       [self.par, {"sim_settings": {"prioritize_top": True,
                                                                                                    "fail_if_placement_issues": True},
                                                                                   "overlap_events": {
                                                                                       "bed": [self.test_overlap_bed, self.test_overlap_bed_2],
                                                                                       "allow_types": "L1"},
                                                                                   "SVs": [{"type": "DEL", "number": 5,
                                                                                            "min_length": 1, "max_length": 5,
                                                                                            "num_overlap": 2}]}],
                                                                       self.hap1, self.hap2, self.bed, self.vcf),
                                            'overlap5': TestProcObject([self.ref_file, {"chr21": "CTCCGTCGTACTAAGTCGTACTCCGTCGTACTAAGTCGTA"}],
                                                                       [self.par, {"sim_settings": {"prioritize_top": True,
                                                                                                    "fail_if_placement_issues": True},
                                                                                   "overlap_events": {
                                                                                        "bed": self.test_overlap_bed_3,
                                                                                        "allow_types": "ALR"},
                                                                                   "SVs": [{"type": "DEL", "number": 5,
                                                                                            "min_length": 1, "max_length": 5,
                                                                                            "num_overlap": 2}]}],
                                                                       self.hap1, self.hap2, self.bed, self.vcf),
                                            'overlap6': TestProcObject([self.ref_file, {"chr21": "CCTCCGTCGTACTAAGTCGTACTAAGTCGTACTCCGTCGTACTAAGTCGTATCCGTCGTACTAAGTCGTACTAAGTCGTACTCCGTCGTACTAAGTCGTA"}],
                                                                       [self.par, {"sim_settings": {"prioritize_top": True,
                                                                                                    "fail_if_placement_issues": True},
                                                                                   "overlap_events": {"bed": self.test_overlap_bed_11,
                                                                                                      "allow_types": ['Alu', 'L1', 'L2', 'SVA', 'HERVK']},
                                                                                   "SVs": [{"type": "DEL", "number": 5,
                                                                                            "min_length": 2, "max_length": 4,
                                                                                            "num_overlap": [1, 1, 1, 1, 1]},
                                                                                           {"type": "DEL", "number": 5,
                                                                                            "min_length": 6, "max_length": 8,
                                                                                            "num_overlap": [1, 1, 1, 1, 1]}]}],
                                                                       self.hap1, self.hap2, self.bed, self.vcf),
                                            'overlap7': TestProcObject([self.ref_file, {"chr21": "CCTCCGTCGTACTAAGTCGTACTAAGTCGTACTCCGTCGTACTAAGTCGTATCCGTCGTACTAAGTCGTACTAAGTCGTACTCCGTCGTACTAAGTCGTA"}],
                                                                       [self.par, {"sim_settings": {"prioritize_top": True,
                                                                                                    "fail_if_placement_issues": True},
                                                                        "overlap_events": {"bed": self.test_overlap_bed_11,
                                                                                           "allow_types": ['Alu', 'L1', 'L2', 'SVA', 'HERVK']},
                                                                                   "SVs": [{"type": "DEL", "number": 5,
                                                                                            "min_length": 1, "max_length": 1,
                                                                                            "num_partial_overlap": [1, 1, 1, 1, 1]},
                                                                                           {"type": "DEL", "number": 5,
                                                                                            "min_length": 2, "max_length": 2,
                                                                                            "num_partial_overlap": [1, 1, 1, 1, 1]}]}],
                                                                       self.hap1, self.hap2, self.bed, self.vcf),
                                            'overlap8': TestProcObject([self.ref_file, {"chr21": "CCTCCGTCGTACTAAGTCGTACTAAGTCGTACTCCGTCGTACTAAGTCGTATCCGTCGTACTAAGTCGTACTAAGTCGTACTCCGTCGTACTAAGTCGTA"}],
                                                                       [self.par, {"sim_settings": {"prioritize_top": True,
                                                                                                    "fail_if_placement_issues": True},
                                                                                   "overlap_events": {"bed": self.test_overlap_bed_11,
                                                                                                      "allow_types": ['Alu', 'L1', 'L2', 'SVA', 'HERVK']},
                                                                                   "SVs": [{"type": "dDUP", "number": 5,
                                                                                            "min_length": [2, 1], "max_length": [4, 1],
                                                                                            "num_overlap": [1, 1, 1, 1, 1]},
                                                                                           {"type": "dDUP", "number": 5,
                                                                                            "min_length": [6, 1], "max_length": [8, 1],
                                                                                            "num_overlap": [1, 1, 1, 1, 1]}]}],
                                                                       self.hap1, self.hap2, self.bed, self.vcf),
                                            'overlap9': TestProcObject([self.ref_file, {"chr21": "CCTCCGTCGTACTAAGTCGTACTAAGTCGTACTCCGTCGTACTAAGTCGTATCCGTCGTACTAAGTCGTACTAAGTCGTACTCCGTCGTACTAAGTCGTA"}],
                                                                       [self.par, {"sim_settings": {"prioritize_top": True,
                                                                                                    "fail_if_placement_issues": True},
                                                                                   "overlap_events": {"bed": self.test_overlap_bed_11,
                                                                                                      "allow_types": ['Alu', 'L1', 'L2', 'SVA', 'HERVK']},
                                                                                   "SVs": [{"type": "dDUP", "number": 5,
                                                                                            "min_length": [1, 1], "max_length": [1, 1],
                                                                                            "num_partial_overlap": [1, 1, 1, 1, 1]},
                                                                                           {"type": "dDUP", "number": 5,
                                                                                            "min_length": [1, 1], "max_length": [2, 1],
                                                                                            "num_partial_overlap": [1, 1, 1, 1, 1]}]}],
                                                                       self.hap1, self.hap2, self.bed, self.vcf)
                                            }
        self.test_objects_alu_mediated = {'alu_med1': TestProcObject([self.ref_file, {"chr21": "CTCCGTCGTACTAAGTCGTACTCCGTCGTACTAAGTCGTA"}],
                                                                     [self.par, {"sim_settings": {"prioritize_top": True,
                                                                                                  "fail_if_placement_issues": True},
                                                                                 "overlap_events": {"bed": self.test_overlap_bed_4},
                                                                                 "SVs": [{"type": "DEL", "number": 1,
                                                                                          "min_length": 13, "max_length": 15,
                                                                                          "num_alu_mediated": 1}]}],
                                                                     self.hap1, self.hap2, self.bed, self.vcf)}

        self.formatter = FormatterIO(self.par)

    def tearDown(self):
        utils.remove_file(self.ins_fasta)
        utils.remove_file(self.bed)
        utils.remove_file(self.vcf)
        utils.remove_file(self.par)

    def initialize_test(self, test_objects_dict, sv_type, output_type='bed', ins_fasta=None):
        # function to execute the shared logic for simulating SVs from test objects and generating bed/vcf output
        config = test_objects_dict[sv_type]
        config.initialize_files()
        curr_sim = SV_Simulator(config.ref, config.par)
        curr_sim.apply_transformations(FastaFile(curr_sim.ref_file))
        if output_type == 'bed':
            if ins_fasta:
                utils.remove_file(ins_fasta)
            self.formatter.export_to_bedpe(curr_sim.svs, self.bed, ins_fasta=ins_fasta)
            records = config.extract_bed_records()
        elif output_type == 'vcf':
            self.formatter.export_to_vcf(curr_sim.svs, curr_sim.stats, vcffile=self.vcf)
            records = config.extract_vcf_records()
        else:
            raise ValueError('output_type must be \'bed\' or \'vcf\'')
        return records

    def extract_ins_fasta(self):
        # extracts single INSSEQ entry from self.ins_fasta (meant for use in isolated INS tests)
        # NB: chrom header output in the form ">chr{number}_{insertion_index}"
        chrom_header, insseq = None, None
        with open(self.ins_fasta) as f:
            for line in f:
                if line[0] == '>':
                    chrom_header = line
                else:
                    insseq = line.rstrip()  # <- strip trailing '\n'
        return chrom_header, insseq

    def singleton_event_bed_tests(self, records, sv_type, chrom, len):
        # helper function to perform the bed file checks shared by all singleton event tests
        self.assertTrue(all([record['source_chr'] == record['target_chr'] == chrom for record in records]))
        self.assertTrue(all([record['parent_type'] == sv_type for record in records]))
        self.assertTrue(all([record['len'] == len for record in records]))
        self.assertTrue(all([record['nth_sv'] == '1' for record in records]))
        self.assertTrue(
            all([record['order'] == str(int(record['ev_type'] in constants.NONZERO_ORDER_OPERATIONS)) for record in
                 records]))
        self.assertTrue(set([record['zyg'] for record in records]) in [{'1/1'}, {'0/1'}, {'1/0'}])

    def singleton_event_vcf_tests(self, record, sv_type, chrom, possible_intervals, possible_targets=None):
        # possible_intervals/targets: list of (start, end) pairs target values the record could take on
        self.assertTrue(record['CHROM'] == chrom)
        self.assertTrue(record['ID'] == record['INFO']['SVTYPE'] == sv_type)
        self.assertIn(record['SAMPLE'], ['1/1', '0/1', '1/0'])
        if sv_type != 'INS':
            self.assertIn((record['POS'], record['INFO']['END']), possible_intervals)
            self.assertTrue(int(record['INFO']['SVLEN']) == (int(record['INFO']['END']) - int(record['POS']) + 1))
        else:
            self.assertIn(record['POS'], [ivl[0] for ivl in possible_intervals])
        if possible_targets:
            self.assertIn(record['INFO']['TARGET'], possible_targets)

    def test_export_bedpe_simple_events(self):
        for sv_type in ['DEL', 'DUP', 'INV', 'INS']:
            record = self.initialize_test(self.test_objects_simple_events, sv_type,
                                          ins_fasta=(self.ins_fasta if sv_type == 'INS' else None))[0]
            self.assertTrue(record['ev_type'] == sv_type)
            self.singleton_event_bed_tests(records=[record], sv_type=sv_type, chrom='chr19', len='3')
            if sv_type == 'INS':
                self.assertTrue((record['source_s'], record['source_e']) in [('0', '0'), ('1', '1')])
                self.assertTrue((record['source_s'], record['source_e']) == (record['target_s'], record['target_e']))
                # test of writing of novel insertion sequences to separate output fasta
                chrom_header, insseq = self.extract_ins_fasta()
                self.assertTrue(chrom_header[1:].split('_')[0] == record['source_chr'])
                self.assertTrue(str(len(insseq)) == record['len'])
            else:
                self.assertTrue(record['source_s'] == record['target_s'] == '0')
                self.assertTrue(record['source_e'] == record['target_e'] == '3')

    def test_export_bedpe_flanked_inversions(self):
        for sv_type in ['dupINVdup', 'delINVdel', 'dupINVdel', 'delINVdup']:
            records = self.initialize_test(self.test_objects_flanked_inversions, sv_type)
            # checks of record fields that will be consistent across all four types
            self.singleton_event_bed_tests(records, sv_type, 'chr19', '2')
            for i in range(2):
                # example events set such that the source events must be placed at positions 0,2,4
                self.assertTrue(records[i]['source_s'] == str(2 * i) and records[i]['source_e'] == str(2 * i + 2))
            # type-wise checks for breakdown under fixed example event
            if sv_type == 'dupINVdup':
                self.assertTrue(records[0]['target_s'] == records[0]['target_e'] == '4')
                self.assertTrue(records[0]['ev_type'] == 'INVDUP')
                self.assertTrue(records[1]['target_s'] == '2')
                self.assertTrue(records[1]['target_e'] == '4')
                self.assertTrue(records[1]['ev_type'] == 'INV')
                self.assertTrue(records[2]['target_s'] == records[2]['target_e'] == '2')
                self.assertTrue(records[2]['ev_type'] == 'INVDUP')
            elif sv_type == 'delINVdel':
                self.assertTrue(all([record['source_s'] == record['target_s'] for record in records]))
                self.assertTrue(all([record['source_e'] == record['target_e'] for record in records]))
                self.assertTrue(records[0]['ev_type'] == 'DEL')
                self.assertTrue(records[1]['ev_type'] == 'INV')
                self.assertTrue(records[2]['ev_type'] == 'DEL')
            elif sv_type == 'dupINVdel':
                self.assertTrue(records[0]['target_s'] == records[0]['target_e'] == '4')
                self.assertTrue(all([record['source_s'] == record['target_s'] for record in records[1:]]))
                self.assertTrue(all([record['source_e'] == record['target_e'] for record in records[1:]]))
                self.assertTrue(records[0]['ev_type'] == 'INVDUP')
                self.assertTrue(records[1]['ev_type'] == 'INV')
                self.assertTrue(records[2]['ev_type'] == 'DEL')
            else:  # <- delINVdup
                self.assertTrue(all([record['source_s'] == record['target_s'] for record in records[:2]]))
                self.assertTrue(all([record['source_e'] == record['target_e'] for record in records[:2]]))
                self.assertTrue(records[2]['target_s'] == records[2]['target_e'] == '2')
                self.assertTrue(records[0]['ev_type'] == 'DEL')
                self.assertTrue(records[1]['ev_type'] == 'INV')
                self.assertTrue(records[2]['ev_type'] == 'INVDUP')

    def test_export_bedpe_dispersions(self):
        for sv_type in ['dDUP', 'INV_dDUP', 'TRA']:
            record = self.initialize_test(self.test_objects_dispersions, sv_type)[0]
            self.singleton_event_bed_tests([record], sv_type, 'chr19', '3')
            # interval checks accounting for forward or backward orientation of dispersion
            self.assertTrue((record['source_s'], record['source_e']) in [('0', '3'), ('3', '6')])
            self.assertTrue((record['target_s'], record['target_e']) in [('0', '0'), ('6', '6')])
            if sv_type == 'dDUP':
                self.assertTrue(record['ev_type'] == 'DUP')
            elif sv_type == 'INV_dDUP':
                self.assertTrue(record['ev_type'] == 'INVDUP')
            else:  # <- TRA
                self.assertTrue(record['ev_type'] == 'TRA')

    def test_export_bedpe_del_inv(self):
        for sv_type in ['delINV', 'INVdel']:
            records = self.initialize_test(self.test_objects_del_inv, sv_type)
            self.singleton_event_bed_tests(records, sv_type, 'chr19', '3')
            self.assertTrue((records[0]['source_s'], records[0]['source_e']) ==
                            (records[0]['target_s'], records[0]['target_e']) == ('0', '3'))
            self.assertTrue((records[1]['source_s'], records[1]['source_e']) ==
                            (records[1]['target_s'], records[1]['target_e']) == ('3', '6'))
            if sv_type == 'delINV':
                self.assertTrue(records[0]['ev_type'] == 'DEL')
                self.assertTrue(records[1]['ev_type'] == 'INV')
            else:  # <- INVdel
                self.assertTrue(records[0]['ev_type'] == 'INV')
                self.assertTrue(records[1]['ev_type'] == 'DEL')

    def test_export_bedpe_idel(self):
        for sv_type in ['dDUP_iDEL', 'INS_iDEL']:
            records = self.initialize_test(self.test_objects_idel, sv_type)
            self.singleton_event_bed_tests(records, sv_type, 'chr19', '3')
            # necessary locations of example event / accounting for forward or backward orientation
            self.assertTrue((records[0]['source_s'], records[0]['source_e']) == ('0', '3'))
            self.assertTrue((records[1]['source_s'], records[1]['source_e']) == ('5', '8'))
            self.assertTrue((records[0]['target_s'], records[0]['target_e']) in [('0', '3'), ('5', '5')])
            self.assertTrue((records[1]['target_s'], records[1]['target_e']) in [('3', '3'), ('5', '8')])
            for record in records:
                # the record with source == target should be the deletion
                if (record['source_s'], record['source_e']) == (record['target_s'], record['target_e']):
                    self.assertTrue(record['ev_type'] == 'DEL')
                else:
                    # for dDUP_iDEL, the source \neq target record is the DUP, for INS_iDEL it is TRA
                    if sv_type == 'dDUP_iDEL':
                        self.assertTrue(record['ev_type'] == 'DUP')
                    else:
                        self.assertTrue(record['ev_type'] == 'TRA')

    def test_export_bedpe_dup_inv(self):
        for sv_type in ['dup_INV', 'INV_dup']:
            records = self.initialize_test(self.test_objects_dup_inv, sv_type)
            self.singleton_event_bed_tests(records, sv_type, 'chr19', '4')
            self.assertTrue((records[0]['source_s'], records[0]['source_e']) == ('0', '4'))
            self.assertTrue((records[1]['source_s'], records[1]['source_e']) == ('4', '8'))
            # for dup_INV the first record will have matching source and target, for INV_dup it will be the second (indexing accordingly)
            self.assertTrue(
                records[int(sv_type == 'dup_INV')]['target_s'] == records[int(sv_type == 'dup_INV')]['source_s'])
            self.assertTrue(
                records[int(sv_type == 'dup_INV')]['target_e'] == records[int(sv_type == 'dup_INV')]['source_e'])
            if sv_type == 'dup_INV':
                self.assertTrue(records[0]['target_s'] == records[0]['target_e'] == '8')
                self.assertTrue(records[0]['ev_type'] == 'INVDUP')
                self.assertTrue(records[1]['ev_type'] == 'INV')
            else:
                self.assertTrue(records[1]['target_s'] == records[1]['target_e'] == '0')
                self.assertTrue(records[0]['ev_type'] == 'INV')
                self.assertTrue(records[1]['ev_type'] == 'INVDUP')

    def test_export_bedpe_INVdup(self):
        record = self.initialize_test(self.test_objects_INVdup, 'INVdup')[0]
        self.singleton_event_bed_tests([record], 'INVdup', 'chr19', '4')
        self.assertTrue(
            (record['source_s'], record['source_e']) == (record['target_s'], record['target_e']) == ('0', '4'))
        self.assertTrue(record['ev_type'] == 'INVDUP')

    def test_nth_sv_entry(self):
        records = self.initialize_test(self.test_objects_multievent, 'INVdup')
        for i in range(len(records)):
            self.assertTrue(records[i]['nth_sv'] == str(i + 1))

    def test_export_vcf_simple_events(self):
        for sv_type in ['DEL', 'DUP', 'INV', 'INS']:
            record = self.initialize_test(self.test_objects_simple_events, sv_type, output_type='vcf')[0]
            possible_intervals = [('1', '3')] if sv_type != 'INS' else [('1', None), ('2', None)]
            self.singleton_event_vcf_tests(record, sv_type, 'chr19', possible_intervals)

    def test_export_vcf_flanked_inversions(self):
        for sv_type in ['dupINVdup', 'delINVdel', 'dupINVdel', 'delINVdup']:
            record = self.initialize_test(self.test_objects_flanked_inversions, sv_type, output_type='vcf')[0]
            self.singleton_event_vcf_tests(record, sv_type, 'chr19', [('1', '6')])

    def test_export_vcf_dispersions(self):
        for sv_type in ['dDUP', 'INV_dDUP', 'TRA']:
            record = self.initialize_test(self.test_objects_dispersions, sv_type, output_type='vcf')[0]
            self.singleton_event_vcf_tests(record, sv_type, 'chr19', [('4', '6'), ('1', '3')], ['0', '6'])

    def test_export_vcf_del_inv(self):
        for sv_type in ['delINV', 'INVdel']:
            record = self.initialize_test(self.test_objects_del_inv, sv_type, output_type='vcf')[0]
            self.singleton_event_vcf_tests(record, sv_type, 'chr19', [('1', '6')])

    def test_export_vcf_idel(self):
        for sv_type in ['dDUP_iDEL', 'INS_iDEL']:
            record = self.initialize_test(self.test_objects_idel, sv_type, output_type='vcf')[0]
            self.singleton_event_vcf_tests(record, sv_type, 'chr19', [('1', '3'), ('6', '8')], ['3', '5'])

    def test_export_vcf_dup_inv(self):
        for sv_type in ['dup_INV', 'INV_dup']:
            record = self.initialize_test(self.test_objects_dup_inv, sv_type, output_type='vcf')[0]
            self.singleton_event_vcf_tests(record, sv_type, 'chr19', [('1', '8')])

    def test_export_vcf_INVdup(self):
        record = self.initialize_test(self.test_objects_INVdup, 'INVdup', output_type='vcf')[0]
        self.singleton_event_vcf_tests(record, 'INVdup', 'chr19', [('1', '4')])

    def test_export_vcf_multievent(self):
        records = self.initialize_test(self.test_objects_multievent, 'INVdup', output_type='vcf')
        for record in records:
            self.assertTrue(record['CHROM'] == 'chr19')
            self.assertTrue(record['ID'] == record['INFO']['SVTYPE'] == 'INVdup')
            self.assertIn(record['SAMPLE'], ['1/1', '0/1', '1/0'])
            self.assertIn(record['INFO']['SVLEN'], ['2', '3', '4'])

    def test_export_overlap(self):
        elt_type_counts = defaultdict(NestedDict(int))
        elt_type_counts['overlap1'] = {'L1HS': 2, 'ALR/Alpha': 1, 'NONE': 2}
        elt_type_counts['overlap2'] = {'L1HS': 2, 'ALR/Alpha': 1, 'NONE': 1}
        # *the vcf INFO overlap event field will reflect the label given in the config (i.e., whether just a prefix or full elt name)
        elt_type_counts['overlap3'] = {'L1': 3, 'ALR': 2}
        elt_type_counts['overlap4'] = {'L1': 2, 'NONE': 3}
        # --> overlap5: case in which allow_types not specified; reported type will take full repName provided in bed file
        elt_type_counts['overlap5'] = {'ALR': 1, 'NONE': 4}
        elt_type_counts['overlap6'] = {'Alu': 1, 'L1': 2, 'L2': 1, 'SVA': 1, 'HERVK': 2, 'NONE': 3}
        elt_type_counts['overlap7'] = {'Alu': 2, 'L1': 2, 'L2': 2, 'SVA': 2, 'HERVK': 2}
        elt_type_counts['overlap8'] = {'Alu': 1, 'L1': 2, 'L2': 1, 'SVA': 1, 'HERVK': 2, 'NONE': 3}
        elt_type_counts['overlap9'] = {'Alu': 2, 'L1': 2, 'L2': 2, 'SVA': 2, 'HERVK': 2}
        elt_type_counts['alu_med1'] = {'ALU_MEDIATED': 1}
        for test_case in ['overlap1', 'overlap2', 'overlap3', 'overlap4', 'overlap5']:
            records = self.initialize_test(self.test_objects_overlap_simple, test_case, output_type='vcf')
            ovlp_evs = [record['INFO']['OVERLAP_EV'] if 'OVERLAP_EV' in record['INFO'].keys() else 'NONE' for record in records]
            self.assertEqual(dict(Counter(ovlp_evs)), elt_type_counts[test_case])
        # Special case tests: low probability failure case when events are failed to be placed (want to repeat until all
        # are placed – we're testing the reporting of the records when all events appear)

        def repeat_until_successful_placement():
            try:
                return self.initialize_test(self.test_objects_overlap_simple, test_case, output_type='vcf')
            except:
                return None

        for test_case in ['overlap6', 'overlap7', 'overlap8', 'overlap9']:
            while True:
                records = repeat_until_successful_placement()
                if records is not None:
                    break
                else:
                    continue
            ovlp_evs = [record['INFO']['OVERLAP_EV'] if 'OVERLAP_EV' in record['INFO'].keys() else 'NONE' for record in records]
            self.assertEqual(dict(Counter(ovlp_evs)), elt_type_counts[test_case])

        for test_case in ['alu_med1']:
            records = self.initialize_test(self.test_objects_alu_mediated, test_case, output_type='vcf')
            ovlp_evs = [record['INFO']['OVERLAP_EV'] if 'OVERLAP_EV' in record['INFO'].keys() else 'NONE' for record in records]
            self.assertEqual(dict(Counter(ovlp_evs)), elt_type_counts[test_case])


if __name__ == "__main__":
    unittest.main()



from insilicosv import utils
import random
from insilicosv.constants import *


class Structural_Variant:
    def __init__(self, sv_type, mode, length_ranges=None, source=None, target=None, vcf_rec=None, ref_fasta=None,
                 overlap_event=None, div_prob=None):
        """
        sv_type: Enum either specifying one of the prewritten classes or a Custom transformation
        mode: flag to indicate whether constructor called in fixed or randomized mode
        length_ranges: list containing tuple(s) (min_length, max_length) or singleton int
        source: tuple representing source sequence, optional
        target: tuple representing target sequence, optional
        vcf_rec: (fixed mode) vcf record giving sv information that will instantiate the event
        ref_fasta: for extracting reference frag for vcf records in fixed mode initialization
        overlap_event: (chr, start, end, elt_type) tuple representing the genome element that this SV is meant to overlap, optional
        """
        self.type = sv_type
        if self.type != Variant_Type.Custom:
            self.source, self.target = SV_KEY[self.type]
            self.name = self.type.name
        else:
            self.source, self.target = tuple(source), tuple(target)
            self.name = str("".join(self.source)) + ">" + str("".join(self.target))

        utils.validate_symbols(self.source, self.target)
        self.source_unique_char, self.target_unique_char = Structural_Variant.reformat_seq(
            self.source), Structural_Variant.reformat_seq(self.target)

        self.start = None
        self.end = None
        self.start_chr = None
        self.req_space = None  # required space for SV, sum of event lengths
        self.source_events = []  # list of Event classes for every symbol in source sequence
        self.events_dict = dict()  # maps every unique symbol in source and target to an Event class
        self.changed_fragments = []  # list recording new fragments to be placed in the output ref
        self.dispersion_flip = False  # orientation of dispersion event
        self.insseq_from_rec = None  # space to store INSSEQ for fixed-mode INS event
        self.overlap_event = overlap_event  # <- element tuple of form (chrom, start, end, type) (optional)
        self.div_prob = div_prob  # <- divergence probability param (optionally given for type == DIVERGENCE)

        if self.type in DISPERSION_TYPES:
            if random.randint(0, 1):
                self.dispersion_flip = True
        if mode == 'randomized':
            self.initialize_events(length_ranges)
        else:
            self.initialize_events_fixed(vcf_rec, ref_fasta)

        self.sv_blocks = Blocks(self)
        self.target_symbol_blocks = self.sv_blocks.target_blocks

        if mode == 'randomized':
            self.active = False
            self.ishomozygous = Zygosity.UNDEFINED
            self.hap = [False, False]
        else:
            # manual call to assign_locations in fixed mode
            self.assign_locations(self.start)

    def __repr__(self):
        return "<SV transformation \"{}\" -> \"{}\" taking up {} non-dispersion spaces>".format(
            ''.join(self.source), ''.join(self.target),
            sum([event.length for event in self.source_events if not event.symbol.startswith(Symbols.DIS.value)]))

    @staticmethod
    def reformat_seq(transformation):
        # transformation: tuple, user inputted source and target
        # if dispersion events exist in transformation, tag on unique ids to make them distinct as they all are "_"
        unique_transform = []
        unique_id = 1
        for component in transformation:
            if component != Symbols.DIS.value and component != Symbols.DUP.value and component != Symbols.DIV.value:
                unique_transform.append(component)
            elif component == Symbols.DUP.value:  # duplication event case, need to group together symbol and duplication marking
                unique_transform[-1] += Symbols.DUP.value
            elif component == Symbols.DIV.value:  # divergence event case, want to keep track of interval needing modification
                unique_transform[-1] += Symbols.DIV.value
            else:  # dispersion event case, component = dispersion
                unique_transform.append(component + str(unique_id))
                unique_id += 1
        return tuple(unique_transform)

    def get_event_frag(self, event, symbol):
        # event: source event from events_dict
        # symbol: target symbol
        decode_funcs = {"invert": lambda string: utils.complement(string[::-1]),
                        "identity": lambda string: string,
                        "complement": utils.complement,
                        "diverge": lambda string: utils.divergence(string,
                                                                   divergence_prob=(1 if self.type == Variant_Type.SNP
                                                                                    else self.div_prob))}
        if any(c.islower() for c in symbol):
            return decode_funcs["invert"](event.source_frag)
        elif symbol[-1] == Symbols.DIV.value:
            return decode_funcs["diverge"](event.source_frag)
        else:
            return event.source_frag

    def initialize_events(self, lengths):
        """
        Initializes event classes and creates a mapping of symbol to event

        lengths: list of tuples specifying min and max length for events within SV
        -> returns list of events in source sequence
        """
        all_symbols = []
        for ele in self.source_unique_char + self.target_unique_char:
            if len(ele) > 0 and (len(ele) == 1 or ele.startswith(Symbols.DIS.value)) and ele.upper() not in all_symbols:
                all_symbols.append(ele.upper())
        all_symbols.sort()

        # symbols_dict: (key = symbol, value = (chosen length, length range))
        symbols_dict = dict()
        if len(lengths) > 1:  # values given by user represents custom ranges for each event symbol of variant in lexicographical order
            assert (len(lengths) == len(all_symbols)), \
                "Number of lengths entered does not match the number of symbols (remember foreign insertions and dispersions) present!"
            for idx, symbol in enumerate(all_symbols):
                symbols_dict[symbol] = (random.randint(lengths[idx][0], lengths[idx][1]), lengths[idx])
        elif len(lengths) == 1:  # value given by user represents length (same range) of each event within variant in lexicographical order
            for symbol in all_symbols:
                symbols_dict[symbol] = (random.randint(lengths[0][0], lengths[0][1]), lengths[0])
        else:
            raise Exception("Lengths parameter expects at least one tuple")

        ovlp_frag = random.choice([frag for frag in self.source_unique_char if frag[0] != '_']) if self.overlap_event is not None else None
        for idx, symbol in enumerate(all_symbols):
            if self.overlap_event is not None and symbol == ovlp_frag:
                ovlp_event_len = int(self.overlap_event[2]) - int(self.overlap_event[1])
                event = Event(self, ovlp_event_len, (ovlp_event_len, ovlp_event_len), symbol,
                              start=int(self.overlap_event[1]), end=int(self.overlap_event[2]))
            else:
                event = Event(self, symbols_dict[symbol][0], symbols_dict[symbol][1], symbol)
            self.events_dict[symbol] = event

        for symbol in self.source_unique_char:
            self.source_events.append(self.events_dict[symbol])

        if self.dispersion_flip:
            self.source_events = self.source_events[::-1]

        self.req_space = sum([event.length for event in self.source_events])

    def initialize_events_fixed(self, vcf_record, ref_fasta):
        # initialization method for SV read in from vcf
        source_len = vcf_record.stop - vcf_record.start if 'SVLEN' not in vcf_record.info else vcf_record.info['SVLEN']
        for symbol in self.source_unique_char:
            if symbol == Symbols.REQUIRED_SOURCE.value:
                source_ev = Event(self, source_len, (source_len, source_len), symbol)
                source_ev.start = vcf_record.start
                source_ev.end = vcf_record.stop
                source_ev.source_chr = vcf_record.chrom
                source_ev.source_frag = ref_fasta.fetch(source_ev.source_chr, source_ev.start, source_ev.end) if \
                    vcf_record.id != Variant_Type.SNP.value else vcf_record.ref
                self.events_dict[symbol] = source_ev
            if symbol.startswith(Symbols.DIS.value):
                self.dispersion_flip = vcf_record.info['TARGET'] < vcf_record.start
                disp_len = vcf_record.info['TARGET'] - vcf_record.stop if not self.dispersion_flip else \
                    vcf_record.start - vcf_record.info['TARGET']
                disp_ev = Event(self, disp_len, (disp_len, disp_len), symbol)
                disp_ev.start = vcf_record.stop if not self.dispersion_flip else vcf_record.info['TARGET']
                disp_ev.end = vcf_record.info['TARGET'] if not self.dispersion_flip else vcf_record.start
                disp_ev.source_chr = vcf_record.chrom
                disp_ev.source_frag = ref_fasta.fetch(disp_ev.source_chr, disp_ev.start, disp_ev.end)
                self.events_dict[symbol] = disp_ev
        # storing the insertion seq for INSs/SNPs with an insertion sequence / alt allele given in the vcf
        if vcf_record.info['SVTYPE'] == 'SNP':
            # NB: only supporting SNP records with a single allele ALT reported
            self.insseq_from_rec = vcf_record.alts[0]
        if vcf_record.info['SVTYPE'] == 'INS':
            if 'INSSEQ' in vcf_record.info:
                self.insseq_from_rec = vcf_record.info['INSSEQ'][0]
            source_ev = Event(self, source_len, (source_len, source_len), Symbols.REQUIRED_SOURCE.value)
            self.events_dict[Symbols.REQUIRED_SOURCE.value] = source_ev
            self.start = vcf_record.start
            self.end = self.start
        else:
            self.start = self.events_dict[Symbols.REQUIRED_SOURCE.value].start
            self.end = self.events_dict[Symbols.REQUIRED_SOURCE.value].end if '_1' not in self.events_dict.keys() else self.events_dict['_1'].end
        self.start_chr = vcf_record.chrom

        # handling for divergent repeat simulation logic (div_dDUPs placed into R1 need to correspond to dDUPs in R2)
        if self.type == Variant_Type.div_dDUP:
            self.target_unique_char = ("A", "_1", "A'")

        self.active = True
        if vcf_record.samples[0]['GT'] == (1, 1):
            self.ishomozygous = Zygosity.HOMOZYGOUS
            self.hap = [True, True]
        else:
            self.ishomozygous = Zygosity.HETEROZYGOUS
            self.hap = random.choice([[True, False], [False, True]])

    def assign_locations(self, start_pos):
        """
        assign events start and end positions (once target blocks are populated and in the right order)
        """
        for block in self.target_symbol_blocks:
            for ev in block:
                ev.source_chr = self.start_chr
                # if the event is one also found in the source, place it at the location given in events_dict
                if ev.symbol.upper() in self.events_dict.keys() or ev.symbol[-1] == Symbols.DIV.value:
                    source_ev = self.events_dict[ev.symbol.upper() if ev.symbol.upper() in self.events_dict.keys() else ev.symbol[0]]
                    ev.start = source_ev.start
                    ev.end = source_ev.end
                    ev.source_frag = self.get_event_frag(source_ev, ev.symbol) if self.insseq_from_rec is None else self.insseq_from_rec

        # everything that wasn't assigned above will be modeled as insertion fragment placed at the nearest event boundary
        target_events = [ev for bl in self.target_symbol_blocks for ev in bl]
        # singleton event that's novel (i.e., INS)
        if len(target_events) == 1 and target_events[0].start is None:
            ev = target_events[0]
            ev.start = start_pos
            ev.end = start_pos
            source_event = self.events_dict[ev.symbol[0].upper()]
            if self.insseq_from_rec is not None:
                ev.source_frag = self.insseq_from_rec
            else:
                ev.source_frag = self.get_event_frag(source_event, ev.symbol)
                if ev.source_frag is None:
                    ev.source_frag = utils.generate_seq(ev.length)
        else:
            for i in range(len(target_events)):
                ev = target_events[i]
                if ev.start is None:
                    if i == 0:
                        # if the first event is novel, set start/end to the start of the nearest event
                        j = i + 1
                        while target_events[j].start is None:
                            j += 1
                        ev.start = target_events[j].start
                        ev.end = target_events[j].start
                    else:
                        ev.start = target_events[i - 1].end
                        ev.end = target_events[i - 1].end
                    source_event = self.events_dict[ev.symbol[0].upper()]
                    ev.source_frag = self.get_event_frag(source_event, ev.symbol)

        self.sv_blocks.generate_target_events_dict()

    def change_fragment(self):
        """
        Takes the mapping of symbols to events and the target sequence to construct a replacement sequence for the reference fragment
        """
        changed_fragments = []
        assert (self.start is not None and self.end is not None), "Undefined SV start for {}".format(
            self)
        block_start = None
        block_end = None

        if self.target_symbol_blocks == [[]]:  # special case: simple deletion -- len(target_symbol_blocks) == 0
            changed_fragments.append([self.start_chr, self.start, self.end, ''])
        else:
            for idx, block in enumerate(self.target_symbol_blocks):
                new_frag = ''
                if len(block) == 0:
                    # TRA branch: delete the A-length interval on the opposite side of the dispersion as our A'
                    if idx == 0:
                        del_len = len(self.target_symbol_blocks[2][0].source_frag)
                        disp_ev = self.target_symbol_blocks[1][0]
                        block_start = disp_ev.start - del_len
                        block_end = disp_ev.start
                    else:
                        del_len = len(self.target_symbol_blocks[idx - 2][0].source_frag)
                        disp_ev = self.target_symbol_blocks[idx - 1][0]
                        block_start = disp_ev.end
                        block_end = block_start + del_len
                    changed_fragments.append([self.start_chr, block_start, block_end, new_frag])
                    continue
                if block[0].symbol.startswith(Symbols.DIS.value):
                    continue
                for i in range(len(block)):
                    ev = block[i]
                    new_frag += ev.source_frag
                    if i == 0:
                        block_start = ev.start
                    if i == len(block) - 1:
                        block_end = ev.end
                changed_fragments.append([self.start_chr, block_start, block_end, new_frag])
            # DEL logic: for all source events whose symbol doesn't appear (in any form) in the target symbols,
            # create a deletion fragment over that interval
            target_symbols = [ev.symbol[0].upper() for bl in self.target_symbol_blocks for ev in bl]
            for source_sym in self.events_dict.keys():
                if not source_sym.startswith(Symbols.DIS.value) and source_sym not in target_symbols:
                    del_ev = self.events_dict[source_sym]
                    changed_fragments.append([del_ev.source_chr, del_ev.start, del_ev.end, ''])

        self.changed_fragments = changed_fragments
        self.clean_event_storage()

    def clean_event_storage(self):
        # remove source fragments from events to save space as they are no longer needed
        for event in self.events_dict.values():
            if event.symbol in self.source_unique_char:  # do not clean out insertion fragments as they'll need to be exported later
                event.source_frag = "Removed"


class Event:
    """
    represents the symbols, also known as the "events," within a SV transformation
    """

    def __init__(self, sv_parent, length, length_range, symbol, source_frag=None, start=None, end=None):
        self.sv_parent = sv_parent  # sv_parent: Structural_Variant, event is always part of larger SV
        self.length = length
        self.length_range = length_range
        self.symbol = symbol  # refers to symbol in SV's transformation
        self.source_chr = None
        self.source_frag = None if not source_frag else source_frag
        self.start = start
        self.end = end

    def __repr__(self):
        return "<Event {}>".format({"length": self.length, "symbol": self.symbol, "start": self.start, "end": self.end,
                                    "source_chr": self.source_chr})


class Blocks:
    """
    Groups together target symbols between dispersion events
    """
    def __init__(self, sv):
        self.sv = sv
        self.target_blocks = []
        self.generate_blocks()
        if self.sv.type in DISPERSION_TYPES and self.sv.dispersion_flip:
            self.flip_blocks()
        self.target_events_dict = None

    def generate_blocks(self):
        self.target_blocks = self.find_blocks(self.sv.target_unique_char)

    def find_blocks(self, transformation):
        # transformation: tuple of strings (source or target chars)
        # -> returns list of lists of events
        # Ex. ("A","B","_","C","D") -> [[Event("A",...),Event("B",...)], [Event("_1")], [Event("C",...),Event("D",...)]]
        # Ex. ("A","B","_","_") -> [[Event("A",...),Event("B",...)],[Event("_1")],[Event("_2")]]
        blocks = [[]]
        for symbol in transformation:
            if symbol.startswith(Symbols.DIS.value):
                source_event = self.sv.events_dict[symbol]
                disp_event = Event(sv_parent=self.sv, length=source_event.length, length_range=None, symbol=symbol)
                blocks.append([disp_event])
                blocks.append([])
            else:
                # used to find corresponding event from encoding, all keys in encoding are in uppercase
                source_event = self.sv.events_dict[symbol[0].upper()]
                target_event = Event(sv_parent=self.sv, length=source_event.length, length_range=None, symbol=symbol)
                # if symbol different from source symbol then being added to input ref
                if symbol.upper() != source_event.symbol:
                    target_event.length = 0
                blocks[-1].append(target_event)
        return blocks

    def flip_blocks(self):
        # perform the optional flip of the blocks list dictated by the flipped dispersion
        self.target_blocks = self.target_blocks[::-1]

    def generate_target_events_dict(self):
        # setter for target_events_dict attribute (to be populated after location assignment of target events)
        self.target_events_dict = {ev.symbol: ev for b in self.target_blocks for ev in b}

    def __repr__(self):
        return f"TARGET_BLOCKS: {self.target_blocks}"




import unittest
import sys
import os

from insilicosv import utils

from insilicosv.simulate import *
from test_simulate import TestObject


class TestKnownSVs(unittest.TestCase):

    def setUp(self):
        self.test_vcf_simple_del = "test/inputs/example_simple_del.vcf"
        self.test_vcf_simple_dup = "test/inputs/example_simple_dup.vcf"
        self.test_vcf_simple_inv = "test/inputs/example_simple_inv.vcf"
        self.test_vcf_simple_ins = "test/inputs/example_simple_ins.vcf"
        self.test_vcf_simple_ins_no_insseq = "test/inputs/example_simple_ins_no_insseq.vcf"
        self.test_vcf_snp = "test/inputs/example_SNP.vcf"
        self.test_vcf_invdup = "test/inputs/example_invdup.vcf"

        self.test_vcf_multidel = "test/inputs/example_multidel.vcf"
        self.test_vcf_del_ins = "test/inputs/example_del_ins.vcf"
        self.test_vcf_del_ins_del = "test/inputs/example_del_ins_del.vcf"
        self.test_vcf_del_dup_del = "test/inputs/example_del_dup_del.vcf"
        self.test_vcf_del_dup = "test/inputs/example_del_dup.vcf"
        self.test_vcf_del_dup_2 = "test/inputs/example_del_dup_2.vcf"
        self.test_vcf_del_inv_del = "test/inputs/example_del_inv_del.vcf"
        self.test_vcf_dup_dup_ins = "test/inputs/example_dup_dup_ins.vcf"
        self.test_vcf_multidel_multisnp = "test/inputs/example_multiDEL_multiSNP.vcf"
        self.test_vcf_div_dDUP = "test/inputs/example_div_dDUP.vcf"
        self.test_vcf_dDUP = "test/inputs/example_dDUP.vcf"
        self.test_vcf_INV_dDUP = "test/inputs/example_INV_dDUP.vcf"
        self.test_vcf_TRA = "test/inputs/example_TRA.vcf"
        self.test_vcf_multi_dispersion = "test/inputs/example_multi_dispersion.vcf"

        self.ref_file = "test/inputs/test.fna"
        self.par = "test/inputs/par.yaml"
        self.hap1 = "test/inputs/test1.fna"
        self.hap2 = "test/inputs/test2.fna"
        self.bed = "test/inputs/out.bed"

        self.test_objects_simple_events = {'DEL': TestObject([self.ref_file, {"chr21": "GCACTATCTCTCCGT"}],
                                                             [self.par, {"SVs": [{"vcf_path": self.test_vcf_simple_del}]}],
                                                             self.hap1, self.hap2, self.bed),
                                           'DUP': TestObject([self.ref_file, {"chr21": "GCACTATCTCTCCGT"}],
                                                             [self.par, {"SVs": [{"vcf_path": self.test_vcf_simple_dup}]}],
                                                             self.hap1, self.hap2, self.bed),
                                           'INV': TestObject([self.ref_file, {"chr21": "GCACTATCTCTCCGT"}],
                                                             [self.par, {"SVs": [{"vcf_path": self.test_vcf_simple_inv}]}],
                                                             self.hap1, self.hap2, self.bed),
                                           'INS': TestObject([self.ref_file, {"chr21": "GCACTATCTCTCCGT"}],
                                                             [self.par, {"SVs": [{"vcf_path": self.test_vcf_simple_ins}]}],
                                                             self.hap1, self.hap2, self.bed),
                                           'INS_2': TestObject([self.ref_file, {"chr21": "GCACTATCTCTCCGT"}],
                                                               [self.par, {"SVs": [{"vcf_path": self.test_vcf_simple_ins_no_insseq}]}],
                                                               self.hap1, self.hap2, self.bed),
                                           'SNP': TestObject([self.ref_file, {"chr21": "GCACTATCTCTCCGT"}],
                                                             [self.par, {"SVs": [{"vcf_path": self.test_vcf_snp}]}],
                                                             self.hap1, self.hap2, self.bed),
                                           'INVdup': TestObject([self.ref_file, {"chr21": "GCACTATCTCTCCGT"}],
                                                                [self.par, {"SVs": [{"vcf_path": self.test_vcf_invdup}]}],
                                                                self.hap1, self.hap2, self.bed),
                                           }
        self.test_objects_multievent = {'multiDEL': TestObject([self.ref_file, {"chr21": "GCACTATCTCTCCGT"}],
                                                               [self.par, {"SVs": [{"vcf_path": self.test_vcf_multidel}]}],
                                                               self.hap1, self.hap2, self.bed),
                                        'del_ins': TestObject([self.ref_file, {"chr21": "GCACTATCTCTCCGT"}],
                                                              [self.par, {"SVs": [{"vcf_path": self.test_vcf_del_ins}]}],
                                                              self.hap1, self.hap2, self.bed),
                                        'del_ins_del': TestObject([self.ref_file, {"chr21": "GCACTATCTCTCCGT"}],
                                                                  [self.par, {"SVs": [{"vcf_path": self.test_vcf_del_ins_del}]}],
                                                                  self.hap1, self.hap2, self.bed),
                                        'del_dup_del': TestObject([self.ref_file, {"chr21": "GCACTATCTCTCCGT"}],
                                                                  [self.par, {"SVs": [{"vcf_path": self.test_vcf_del_dup_del}]}],
                                                                  self.hap1, self.hap2, self.bed),
                                        'del_inv_del': TestObject([self.ref_file, {"chr21": "GCACTATCTCTCCGT"}],
                                                                  [self.par, {"SVs": [{"vcf_path": self.test_vcf_del_inv_del}]}],
                                                                  self.hap1, self.hap2, self.bed),
                                        'dup_dup_ins': TestObject([self.ref_file, {"chr21": "GCACTATCTCTCCGT"}],
                                                                  [self.par, {"SVs": [{"vcf_path": self.test_vcf_dup_dup_ins}]}],
                                                                  self.hap1, self.hap2, self.bed),
                                        'multiDEL_multiSNP': TestObject([self.ref_file, {"chr21": "GCACTATCTCTCCGT"}],
                                                                        [self.par, {"SVs": [{"vcf_path": self.test_vcf_multidel_multisnp}]}],
                                                                        self.hap1, self.hap2, self.bed),
                                        }
        self.test_objects_dispersions = {'div_dDUP': TestObject([self.ref_file, {"chr21": "GCACTATCTCTCCGT"}],
                                                                [self.par, {"SVs": [{"vcf_path": self.test_vcf_div_dDUP}]}],
                                                                self.hap1, self.hap2, self.bed),
                                         'dDUP': TestObject([self.ref_file, {"chr21": "GCACTATCTCTCCGT"}],
                                                            [self.par, {"SVs": [{"vcf_path": self.test_vcf_dDUP}]}],
                                                            self.hap1, self.hap2, self.bed),
                                         'INV_dDUP': TestObject([self.ref_file, {"chr21": "GCACTATCTCTCCGT"}],
                                                                [self.par, {"SVs": [{"vcf_path": self.test_vcf_INV_dDUP}]}],
                                                                self.hap1, self.hap2, self.bed),
                                         'TRA': TestObject([self.ref_file, {"chr21": "GCACTATCTCTCCGT"}],
                                                           [self.par, {"SVs": [{"vcf_path": self.test_vcf_TRA}]}],
                                                           self.hap1, self.hap2, self.bed),
                                         'multievent': TestObject([self.ref_file, {"chr21": "GCACTATCTCTCCGT"}],
                                                                  [self.par,
                                                                   {"SVs": [{"vcf_path": self.test_vcf_multi_dispersion}]}],
                                                                  self.hap1, self.hap2, self.bed)}

    def helper_test_simple_sv(self, config_event_obj, target_frags=None):
        # template test method for simple SVs
        config = config_event_obj
        config.initialize_files()
        fixed_sim = SV_Simulator(config.ref, config.par)
        fixed_sim.produce_variant_genome(config.hap1, config.hap2, config.ref, config.bed, export_to_file=False)
        changed_frag_hap1, changed_frag_hap2 = config.get_actual_frag(return_haps='both')
        config.remove_test_files()
        if target_frags is not None:
            self.assertTrue(changed_frag_hap1 in target_frags or changed_frag_hap2 in target_frags)
        return changed_frag_hap1, changed_frag_hap2

    def test_simple_events(self):
        self.helper_test_simple_sv(self.test_objects_simple_events['DEL'], ['GCCTCCGT'])
        self.helper_test_simple_sv(self.test_objects_simple_events['DUP'], ['GCACTATCTACTATCTCTCCGT'])
        self.helper_test_simple_sv(self.test_objects_simple_events['INV'], ['GCAGATAGTCTCCGT'])
        self.helper_test_simple_sv(self.test_objects_simple_events['INS'], ['GCGGGGGGGACTATCTCTCCGT'])
        # insertion without specified INSSEQ (insertion sequence randomly generated)
        frag1, frag2 = self.helper_test_simple_sv(self.test_objects_simple_events['INS_2'])
        self.assertTrue(len(frag1) == 22 and len(frag2) == 22)
        self.helper_test_simple_sv(self.test_objects_simple_events['SNP'], ['GCGCTATCTCTCCGT'])
        self.helper_test_simple_sv(self.test_objects_simple_events['INVdup'], ['GCAGATAGTAGATAGTCTCCGT'])

    def test_multiple_events(self):
        # both DELs heterozygous
        self.helper_test_simple_sv(self.test_objects_multievent['multiDEL'], ['GCTATCTCCGT'])
        self.helper_test_simple_sv(self.test_objects_multievent['multiDEL_multiSNP'], ['AATATCTCCGT'])
        # hom. DEL, het. INS
        self.helper_test_simple_sv(self.test_objects_multievent['del_ins'], ['GCGGGGGGGACTATCTCGT', 'GCACTATCTCGT'])
        # het. DELs, hom. INS
        self.helper_test_simple_sv(self.test_objects_multievent['del_ins_del'],
                                   ['GCGGGGGGGTATCTCGT', 'GCACGGGGGGGTATCTCGT',
                                    'GCGGGGGGGTATCTCTCCGT', 'GCACGGGGGGGTATCTCTCCGT'])
        self.helper_test_simple_sv(self.test_objects_multievent['del_dup_del'], ['GCTATCTATCTCGT', 'GCACTATCTATCTCGT',
                                                                                 'GCTATCTCGT', 'GCACTATCTCGT'])
        self.helper_test_simple_sv(self.test_objects_multievent['del_inv_del'], ['GCGATATCGT'])
        # het., hom., het.
        self.helper_test_simple_sv(self.test_objects_multievent['dup_dup_ins'], ['GCACACGGGGGGGTATCTCTCCTCCGT',
                                                                                 'GCACACTATCTCTCCTCCGT',
                                                                                 'GCACGGGGGTATCTCTCCTCCGT'])

    def test_dispersion_events(self):
        self.helper_test_simple_sv(self.test_objects_dispersions['div_dDUP'], ['GCACTATCTCACTATTCCGT'])
        self.helper_test_simple_sv(self.test_objects_dispersions['dDUP'], ['GCACTATCTCACTATTCCGT'])
        self.helper_test_simple_sv(self.test_objects_dispersions['INV_dDUP'], ['GCACTATCTCATAGTTCCGT'])
        self.helper_test_simple_sv(self.test_objects_dispersions['TRA'], ['GCCTCACTATTCCGT'])
        self.helper_test_simple_sv(self.test_objects_dispersions['multievent'], ['GACGGCCTACTATCACTATTCCGT'])

    def test_get_original_pos(self):
        for i in range(20):
            self.assertTrue(utils.get_original_base_pos(self.test_vcf_simple_del, i, 'chr21') == i + (7 if i >= 2 else 0))
            self.assertTrue(utils.get_original_base_pos(self.test_vcf_simple_dup, i, 'chr21') == i - (7 if i >= 9 else 0))
            self.assertTrue(utils.get_original_base_pos(self.test_vcf_simple_inv, i, 'chr21') == (i if i < 2 or i >= 9 else 10 - i))
            self.assertTrue(utils.get_original_base_pos(self.test_vcf_del_dup, i, 'chr21') == i + (3 if 1 < i < 11 else 0))
        self.assertTrue(utils.get_original_base_pos(self.test_vcf_multidel, 1, 'chr21'), 1)
        self.assertTrue(utils.get_original_base_pos(self.test_vcf_multidel, 9, 'chr21'), 11)
        self.assertTrue(utils.get_original_base_pos(self.test_vcf_multidel, 13, 'chr21'), 17)
        for i in range(20):
            self.assertTrue(utils.get_original_base_pos(self.test_vcf_del_dup_2, i, 'chr21') == i + 2 * (i > 0) - 4 * (i > 5))
            self.assertTrue(utils.get_original_base_pos(self.test_vcf_del_dup_del, i, 'chr21') == i + 2 * (i > 0) - 4 * (i > 5) + 3 * (i > 10))
            self.assertTrue(utils.get_original_base_pos(self.test_vcf_del_inv_del, i, 'chr21') == ((9 - i) if 1 < i < 6 else (i + 2 * (i > 0) + 3 * (i > 6))))


if __name__ == '__main__':
    unittest.main()



import yaml
import sys
import pysam
import argparse
import os
import time
from insilicosv import utils
from insilicosv.constants import *
from pysam import FastaFile


class FormatterIO:
    def __init__(self, par_file):
        self.bedpe_counter = 1
        self.par_file = par_file
        self.config = None

    @staticmethod
    def run_checks_randomized(config):
        """
        check method for yaml given with SVs given for randomized placement on reference
        """
        config_svs = config['SVs']
        for config_sv in config_svs:
            if "avoid_intervals" in config_sv:
                continue
            elif "type" not in config_sv:
                raise Exception("\"Type\" attribute must be specified! For custom transformations, enter in \"Custom\"")
            elif config_sv["type"] == "SNP":  # SNP events are only specified by count (size is deterministic)
                if "number" in config_sv:
                    continue
                else:
                    raise Exception("Number is a required parameter for all SVs")
            # makes sure required attributes are written into parameter file
            if "min_length" not in config_sv:
                raise Exception("Min length must be specified on all SVs!")
            if "max_length" not in config_sv:
                raise Exception("Max length must be specified on all SVs!")
            if "number" not in config_sv:
                raise Exception("Number is a required parameter for all SVs")

            # makes sure attributes are the correct datatype
            elif "type" in config_sv and not isinstance(config_sv["type"], str):
                raise Exception("Invalid {} type for SV \'type\' attribute, str expected".format(type(config_sv["type"])))
        valid_optional_par = ["fail_if_placement_issues", "max_tries", "generate_log_file", "filter_small_chr",
                              "prioritize_top", "homozygous_only"]  # valid arguments within sim_settings
        for parameter in config['sim_settings']:
            if parameter not in valid_optional_par:
                raise Exception("\"{}\" is an invalid argument under sim_settings".format(parameter))
        valid_keys = ["sim_settings", "SVs", "overlap_events"]  # valid arguments at the top level
        for key in config:
            if key not in valid_keys:
                raise Exception("Unknown argument \"{}\"".format(key))

    def postproc_config_dict(self):
        if "vcf_path" not in self.config["SVs"][0]:
            self.run_checks_randomized(self.config)
        for config_sv in self.config['SVs']:
            if "avoid_intervals" in config_sv or "vcf_path" in config_sv:
                continue
            # SV event length specification - not applicable for SNPs
            if config_sv["type"] != "SNP":
                if isinstance(config_sv["min_length"], int):
                    config_sv["length_ranges"] = [(config_sv["min_length"], config_sv["max_length"])]
                else:
                    config_sv["length_ranges"] = list(zip(config_sv["min_length"], config_sv["max_length"]))
                # make sure max_length >= min_length >= 0
                assert all(max_len >= min_len >= 0 for (min_len, max_len) in config_sv["length_ranges"]), "Max length must be >= min length for all SVs! Also ensure that all length values are >= 0."

            config_sv["type"] = Variant_Type(config_sv["type"])
            if config_sv["type"] != Variant_Type.Custom:
                config_sv["source"] = None
                config_sv["target"] = None

        # setting default values for sim_settings fields
        if 'sim_settings' not in self.config.keys():
            self.config['sim_settings'] = {'max_tries': 50, 'prioritize_top': True, 'fail_if_placement_issues': False}
        else:
            if 'max_tries' not in self.config['sim_settings']:
                self.config['sim_settings']['max_tries'] = 50
            if 'fail_if_placement_issues' not in self.config['sim_settings']:
                self.config['sim_settings']['fail_if_placement_issues'] = False

    def yaml_to_var_list(self):
        try:
            with open(self.par_file) as yaml_file:
                self.config = yaml.full_load(yaml_file)
        except:
            raise Exception("YAML File {} failed to be open".format(self.par_file))
        self.postproc_config_dict()

    def write_to_file(self, sv, bedfile, source_s, source_e, target_s, target_e, transform, event, nth_sv, order=0):
        assert (not event.symbol.startswith(Symbols.DIS.value))
        if transform == Operations.INS.value:
            transform_length = event.length
        else:
            transform_length = source_e - source_s
        if event.length > 0:
            with open(bedfile, "a") as fout:
                row = [str(event.source_chr),
                       str(source_s),
                       str(source_e),
                       str(event.source_chr),
                       str(target_s),
                       str(target_e),
                       transform,
                       str(transform_length),
                       '%d/%d' % (int(sv.hap[0]), int(sv.hap[1])),
                       sv.name,
                       str(nth_sv),
                       str(order)]
                fout.write("\t".join(row) + "\n")

    @staticmethod
    def symbol_is_inversion(symbol):
        return any(c.islower() for c in symbol)

    @staticmethod
    def export_insertions(chr, start_pos, seq, ins_fasta):
        """
        Exports foreign insertion sequences to separate fasta file, append only
        """
        with open(ins_fasta, "a") as fout_ins:
            fout_ins.write(">{}_{}\n".format(chr, start_pos))
            fout_ins.write("{}\n".format(seq))

    @staticmethod
    def get_event_target_operation(ev, target_events_dict, source_events_dict):
        """
        determines target interval and operation for multi-source events
        """
        # A -> A'
        if ev + Symbols.DUP.value in target_events_dict.keys():
            trg_sym = ev + Symbols.DUP.value
            return (target_events_dict[trg_sym].start, target_events_dict[trg_sym].end), \
                Operations.DUP.value if ev in target_events_dict.keys() else Operations.TRA.value
        # A -> a'
        elif ev.lower() + Symbols.DUP.value in target_events_dict.keys():
            trg_sym = ev.lower() + Symbols.DUP.value
            return (target_events_dict[trg_sym].start, target_events_dict[trg_sym].end), Operations.INVDUP.value
        # A -> a
        elif ev.lower() in target_events_dict.keys():
            trg_sym = ev.lower()
            return (target_events_dict[trg_sym].start, target_events_dict[trg_sym].end), Operations.INV.value
        # A -> A* (in the case of a custom event in which an event is divergently duplicated)
        elif ev + Symbols.DIV.value in target_events_dict.keys():
            trg_sym = ev + Symbols.DIV.value
            return (target_events_dict[trg_sym].start, target_events_dict[trg_sym].end), Operations.DIV.value
        # A -> A (insertion if source A is undefined, identity otherwise)
        elif ev in target_events_dict.keys():
            return (target_events_dict[ev].start, target_events_dict[ev].end), \
                Operations.INS.value if source_events_dict[ev].start is None else Operations.IDENTITY.value
        # A -> [none]
        elif ev not in [sym[0] for sym in target_events_dict.keys()]:
            return (source_events_dict[ev].start, source_events_dict[ev].end), Operations.DEL.value
        # otherwise unknown mapping
        else:
            return (source_events_dict[ev].start, source_events_dict[ev].end), Operations.UNDEFINED.value

    @staticmethod
    def postprocess_record_params(sv, sv_record_info):
        """
        arrange the bed_record parameter dictionaries in order of ascending source interval start position
        and assign order values to the relevant entries
        """
        # for TRA/INS/DUP events with the same target position, 'order' describes the order in which they
        # are compiled (i.e., the order in which they appear in the target sequence)
        order = 0
        ins_pos = None
        for block in sv.target_symbol_blocks:
            for target_event in block:
                if target_event.symbol.startswith(Symbols.DIS.value) or \
                        target_event.symbol in sv_record_info.keys():  # <- prevent collision with A' and A if both in target
                    continue
                src_sym = target_event.symbol[0].upper()
                if sv_record_info[src_sym]['transform'] in NONZERO_ORDER_OPERATIONS:
                    if ins_pos is None:
                        ins_pos = sv_record_info[src_sym]['target_s']
                        order += 1
                    elif sv_record_info[src_sym]['target_s'] == ins_pos:
                        order += 1
                else:
                    ins_pos = None
                    order = 0
                sv_record_info[src_sym]['order'] = order
        return sorted([params for params in sv_record_info.values()], key=lambda params: params['source_s'])

    def export_to_bedpe(self, svs, bedfile, ins_fasta=None, reset_file=True):
        if reset_file:
            utils.reset_file(bedfile)
            if ins_fasta:
                utils.reset_file(ins_fasta)
        for nth_sv, sv in enumerate(svs):
            # SVs with multiple source events will be split into multiple bed records (one for each)
            if len(sv.events_dict) == 1:
                ev = list(sv.sv_blocks.target_events_dict.values())[0] if sv.type == Variant_Type.INS\
                        else list(sv.events_dict.values())[0]
                op = self.get_event_target_operation(ev.symbol, sv.sv_blocks.target_events_dict, sv.events_dict)[1]
                record_info = {'source_s': ev.start, 'source_e': ev.end, 'target_s': ev.start, 'target_e': ev.end,
                               'transform': op, 'sv': sv, 'event': ev, 'bedfile': bedfile, 'nth_sv': nth_sv + 1,
                               'order': int(op in NONZERO_ORDER_OPERATIONS)}
                self.write_to_file(**record_info)
                if op == Operations.INS.value:
                    self.export_insertions(sv.start_chr, ev.start, ev.source_frag, ins_fasta)
            else:
                # multiple source events: source intervals taken from the source events
                # and target intervals taken from corresponding target events (if no match, then deletion)
                sv_record_info = {}
                for ev in sv.events_dict.values():
                    # skip dispersion events
                    if ev.symbol.startswith(Symbols.DIS.value):
                        continue
                    sv_record_info[ev.symbol] = {'source_s': ev.start, 'source_e': ev.end, 'sv': sv, 'event': ev, 'bedfile': bedfile, 'nth_sv': nth_sv + 1}
                    (target_s, target_e), operation = self.get_event_target_operation(ev.symbol, sv.sv_blocks.target_events_dict, sv.events_dict)
                    sv_record_info[ev.symbol]['target_s'] = target_s
                    sv_record_info[ev.symbol]['target_e'] = target_e
                    sv_record_info[ev.symbol]['transform'] = operation
                for param_dict in self.postprocess_record_params(sv, sv_record_info):
                    self.write_to_file(**param_dict)

    def export_to_vcf(self, svs, stats, vcffile):
        with open(vcffile, "w") as vcf:
            vcf.write("##fileformat=VCFv4.2\n")
            for chrm, chrm_len in stats.chr_lengths.items():
                vcf.write("##contig=<ID=%s,length=%d>\n" % (chrm, chrm_len))
            vcf.write("#%s\n" % "\t".join(["CHROM", "POS", "ID", "REF", "ALT", "QUAL", "FILTER", "INFO", "FORMAT",
                                           "SAMPLE"]))
        # *** This will throw an error with pysam version 0.18, need 0.16.0.1
        vcf_file = pysam.VariantFile(vcffile)
        vcf_file.header.info.add('END', number=1, type='Integer', description="End position of the variant "
                                                                              "described in this record")
        vcf_file.header.info.add('CIPOS', number=2, type='Integer', description="Confidence interval around POS for "
                                                                                "imprecise variants")
        vcf_file.header.info.add('CIEND', number=2, type='Integer', description="Confidence interval around END for "
                                                                                "imprecise variants")
        vcf_file.header.info.add('SVTYPE', number=1, type='String', description="Type of structural variant")
        vcf_file.header.info.add('SVLEN', number=1, type='Integer', description="Length of structural variant")
        vcf_file.header.info.add('SVMETHOD', number=1, type='String', description="SV detection method")
        vcf_file.header.info.add('TARGET', number=1, type='Integer', description="Target location for divergent repeat")
        vcf_file.header.info.add('OVERLAP_EV', number=1, type='String', description="Bool. indicator for the event being"
                                                                                    "placed at an overlap_events interval")
        vcf_file.header.formats.add('GT', number=1, type='String', description="Genotype")

        vcf_out_file = pysam.VariantFile(vcffile, 'w', header=vcf_file.header)

        for sv in svs:
            zyg = (int(sv.hap[0]), int(sv.hap[1]))
            dispersion_target = None
            if sv.type in DISPERSION_TYPES:
                source_event = sv.events_dict[Symbols.REQUIRED_SOURCE.value]
                disp_event = sv.events_dict['_1']
                rec_start = source_event.start
                rec_end = source_event.end
                if disp_event.start == source_event.end:
                    dispersion_target = disp_event.end
                else:
                    dispersion_target = disp_event.start
            else:
                rec_start = min([frag[1] for frag in sv.changed_fragments])
                rec_end = max(frag[2] for frag in sv.changed_fragments)
            if dispersion_target is not None:
                info_field = {'SVTYPE': sv.type.value, 'SVLEN': rec_end - rec_start, 'TARGET': dispersion_target}
            else:
                if sv.type == Variant_Type.INS:
                    # special case of simple INS: sv length \neq (sv end - sv start)
                    # **pysam will delete END fields that are equal to POS, therefore INS records won't have an END
                    rec_end += 1
                    info_field = {'SVTYPE': sv.type.value, 'SVLEN': sv.events_dict[Symbols.REQUIRED_SOURCE.value].length}
                else:
                    info_field = {'SVTYPE': sv.type.value, 'SVLEN': rec_end - rec_start}
            if sv.overlap_event is not None:
                info_field['OVERLAP_EV'] = sv.overlap_event[3]

            vcf_record = vcf_out_file.header.new_record(contig=sv.start_chr, start=rec_start, stop=rec_end,
                                                        alleles=['N', '<%s>' % sv.type.value], id=sv.type.value,
                                                        info=info_field,
                                                        qual=100, filter='PASS',
                                                        samples=[{'GT': zyg}])
            vcf_out_file.write(vcf_record)

        vcf_out_file.close()

    def export_variants_to_fasta(self, id, edits, fasta_out, fasta_file, verbose=False):
        """
        Exports list of changes from simulator to fasta file

        id: chr_id to apply edits to
        edits: list with elements of the form (start, end, new_frag)
        fasta_out: Fasta file to export changes to
        fasta_file: FastaFile with access to reference
        """
        with open(fasta_out, "a") as fout_export:
            if id not in fasta_file.references:
                raise KeyError("ID {} not found in inputted fasta file".format(id))
            if verbose:
                print("New ID: ", id)
            fout_export.write(">" + str(id) + "\n")
            chr_variants = list(edits)
            chr_variants.sort()
            chr_variants.append([fasta_file.get_reference_length(id), fasta_file.get_reference_length(id), ""])
            pos = 0
            for variant in chr_variants:
                var_start, var_end = variant[0], variant[1]
                while pos < var_start:
                    appropriate_buffer = MAX_BUFFER_SIZE if var_start - pos > MAX_BUFFER_SIZE else var_start - pos
                    c = fasta_file.fetch(id, pos, pos + appropriate_buffer)
                    fout_export.write(c)
                    pos += appropriate_buffer
                assert (pos == var_start), "Replacement fragment about to be inserted at position {} instead of var_start {}".format(pos, var_start)
                fout_export.write(variant[2])
                pos = var_end
            fout_export.write("\n")

    def close(self):
        self.fin_export1.close()
        self.fin_export2.close()


def collect_args():
    parser = argparse.ArgumentParser(description='insilicoSV is a software to design and simulate complex structural variants, both novel and known.')
    parser.add_argument("ref", help="FASTA reference file")
    parser.add_argument("config", help="YAML config file")
    parser.add_argument("prefix", help="Used for naming all output files (haplotype, bed, and stats)")
    parser.add_argument("-r", "--root", action="store", metavar="DIR", dest="root_dir", help="root directory for all files given")
    parser.add_argument("--random_seed", type=int,
                        help="if non-zero, random seed for random number generation")

    args = parser.parse_args()
    args_dict = {"ref": args.ref, "config": args.config, "ins_fasta": args.prefix + ".insertions.fa", "hap1": args.prefix + ".hapA.fa",
                 "hap2": args.prefix + ".hapB.fa", "bedpe": args.prefix + ".bed", "stats": args.prefix + ".stats.txt", "log_file": args.prefix + ".log",
                 "random_seed": args.random_seed}
    if args.root_dir:
        for key, curr_path in args_dict.items():
            args_dict[key] = os.path.join(args.root_dir, curr_path)
    return args_dict





import random
from insilicosv import utils
import logging
import time
from os import write
from insilicosv.processing import FormatterIO, collect_args
from pysam import FastaFile
from pysam import VariantFile
from insilicosv.constants import *
from insilicosv.structural_variant import Structural_Variant, Event
from collections import defaultdict

time_start = time.time()


class StatsCollection:
    """collection of information for stats file, if requested"""
    def __init__(self, chr_ids, chr_lens):
        self.num_heterozygous = 0
        self.num_homozygous = 0
        self.total_svs = 0
        self.active_svs = 0
        self.active_events_chr = dict()
        self.chr_ids = chr_ids
        self.chr_lengths = chr_lens
        self.avg_len = [0, 0]  # Average length of SV events/components
        self.len_frags_chr = dict()  # Lengths of altered fragments within chromosome
        self.sv_types = dict()

        for id in self.chr_ids:
            self.len_frags_chr[id] = 0
            self.active_events_chr[id] = 0

    def get_info(self, svs):
        """
        collects all information for stats file after all edits are completed

        svs: list of Structural_Variant objects
        -> return None
        """
        self.total_svs = len(svs)
        self.min_event_len = min([event.length for sv in svs if sv.active for key, event in sv.events_dict.items() if
                                  not event.symbol.startswith(Symbols.DIS.value)])
        self.max_event_len = max([event.length for sv in svs if sv.active for key, event in sv.events_dict.items() if
                                  not event.symbol.startswith(Symbols.DIS.value)])

        for sv in svs:
            if sv.active:  # only collect information for SVs that were successfully simulated
                self.active_svs += 1
                # count up SVs of each type
                if sv.name in self.sv_types:
                    self.sv_types[sv.name] += 1
                else:
                    self.sv_types[sv.name] = 1

                # count up zygosity
                if sv.hap[0] and sv.hap[1]:  # homozygous SV
                    self.num_homozygous += 1
                else:  # heterozygous SV
                    self.num_heterozygous += 1

                # add up the lengths of impacted regions on the reference
                for frag in sv.changed_fragments:
                    self.len_frags_chr[frag[0]] += frag[2] - frag[1]

                # count up average length of non-dispersion events
                for symbol in sv.events_dict:
                    if not symbol.startswith(Symbols.DIS.value):
                        event = sv.events_dict[symbol]
                        self.avg_len[0] += event.length
                        self.avg_len[1] += 1
        self.avg_len = self.avg_len[0] // self.avg_len[1] if self.avg_len[1] != 0 else 0

    def export_data(self, fileout):
        """
        Exports all collected data to entered file
        fileout: Location to export stats file
        """
        def write_item(fout, name, item, prefix=""):
            fout.write("{}{}: {}\n".format(prefix, str(name), str(item)))

        with open(fileout, "w") as fout:
            fout.write("===== Overview =====\n")
            write_item(fout, "SVs successfully simulated", str(self.active_svs) + "/" + str(self.total_svs))
            for sv_type in self.sv_types:
                write_item(fout, sv_type, self.sv_types[sv_type], prefix="\t- ")
            write_item(fout, "Homozygous SVs", self.num_homozygous)
            write_item(fout, "Heterozygous SVs", self.num_heterozygous)
            write_item(fout, "Average length of SV symbols/components (excluding dispersions)", self.avg_len)
            write_item(fout, "Min length of non-dispersion event", self.min_event_len)
            write_item(fout, "Max length of non-dispersion event", self.max_event_len)
            for id in self.chr_ids:
                fout.write("\n===== {} =====\n".format(id))
                write_item(fout, "Length of sequence", self.chr_lengths[id])
                write_item(fout, "Total impacted length of reference chromosome", self.len_frags_chr[id])


class SV_Simulator:
    def __init__(self, ref_file, par_file, log_file=None):
        """
        ref_file: file location to reference genome (.fasta or .fna or .fa)
        par_file: file location to configuration file (.yaml)
        log_file: location to store log file with diagnostic information if config parameters indicate so
        """
        global time_start
        print("Setting up Simulator...")

        self.ref_file = ref_file
        self.ref_fasta = FastaFile(ref_file)  # FastaFile generates a new index file if one is not found
        self.formatter = FormatterIO(par_file)
        self.formatter.yaml_to_var_list()
        config = self.formatter.config
        self.svs_config = config['SVs']  # config for what SVs to simulate

        self.sim_settings = config['sim_settings']
        if log_file and "generate_log_file" in self.sim_settings.keys():
            logging.basicConfig(filename=log_file, filemode="w", level=logging.DEBUG,
                                format='[%(name)s: %(levelname)s - %(asctime)s] %(message)s')
            self.log_to_file("YAML Configuration: {}".format(config))

        # get all chromosome ids
        self.order_ids = self.ref_fasta.references
        self.len_dict = dict()  # stores mapping with key = chromosome, value = chromosome length
        for id in self.order_ids:
            chrom_len = self.ref_fasta.get_reference_length(id)
            if 'filter_small_chr' in self.sim_settings and chrom_len < self.sim_settings['filter_small_chr']:
                print("Filtering chromosome {}: Length of {} below threshold of {}".format(id, chrom_len, self.sim_settings['filter_small_chr']))
            else:
                self.len_dict[id] = chrom_len
                print("Length of chromosome {}: {}".format(id, self.len_dict[id]))

        # initialize stats file to be generated after all edits and exporting are finished
        self.stats = StatsCollection(self.order_ids, self.len_dict)

        self.mode = "randomized"
        self.vcf_path = None
        if "vcf_path" in self.svs_config[0]:
            self.mode = "fixed"
            self.vcf_path = self.svs_config[0]["vcf_path"]

        self.svs = []
        self.event_ranges = defaultdict(list)

        for d in self.svs_config:
            if "avoid_intervals" in d:
                # extract {chrom: [(start, end)]} intervals from vcf, add intervals from vcf to event range
                self.extract_vcf_event_intervals(d["avoid_intervals"])

        self.overlap_events = None if "overlap_events" not in config.keys() \
            else utils.OverlapEvents(config, allow_chroms=self.order_ids)

        self.initialize_svs()

        print("Finished Setting up Simulator in {} seconds\n".format(time.time() - time_start))
        time_start = time.time()

    def __repr__(self):
        return "All structural variants entered into simulator: {}".format(self.svs)

    def log_to_file(self, info, key="DEBUG"):
        # only logs to file if config setting indicates so
        log_func = [logging.debug, logging.warning]
        key_to_func = {"DEBUG": logging.debug, "WARNING": logging.warning}
        if "generate_log_file" in self.sim_settings and self.sim_settings["generate_log_file"]:
            key_to_func[key](info)
            # logging.debug(info)

    def get_rand_chr(self, check_size=None, fixed_chrom=None):
        # random assignment of SV to a chromosome (unless we have a predetermined chromosome for this event)
        # -> returns random chromosome and its length
        valid_chrs = self.order_ids
        if check_size is not None:
            valid_chrs = [chrom for chrom, chr_size in self.len_dict.items() if chr_size >= check_size]
        if len(valid_chrs) == 0:
            raise Exception("SVs are too big for the reference!")
        rand_id = valid_chrs[random.randint(0, len(valid_chrs) - 1)] if fixed_chrom is None else fixed_chrom
        chr_len = self.len_dict[rand_id]
        chr_event_ranges = self.event_ranges[rand_id]
        assert rand_id is not None
        return rand_id, chr_len, chr_event_ranges

    def extract_vcf_event_intervals(self, vcf_path):
        vcf = VariantFile(vcf_path)
        for rec in vcf.fetch():
            self.event_ranges[rec.chrom].append((rec.start, rec.stop))

    def process_vcf(self, vcf_path):
        # process vcf containing SVs to be added (deterministically) to reference
        active_svs_total = 0
        time_start_local = 0
        vcf = VariantFile(vcf_path)
        for rec in vcf.fetch():
            svtype = Variant_Type(rec.info['SVTYPE']) if 'SVTYPE' in rec.info else Variant_Type(rec.id)
            self.event_ranges[rec.chrom].append((rec.start, rec.stop))
            sv = Structural_Variant(sv_type=svtype, mode='fixed', vcf_rec=rec, ref_fasta=self.ref_fasta)
            self.svs.append(sv)
            active_svs_total += 1
            self.log_to_file("Intervals {} added to Chromosome \"{}\"".format(self.event_ranges[rec.chrom], rec.chrom))
            time_dif = time.time() - time_start_local
            print("{} SVs successfully placed ========== {} seconds".format(active_svs_total, time_dif), end="\r")
            time_start_local = time.time()

    def initialize_svs(self):
        """
        Creates Structural_Variant objects for every SV to simulate and decides zygosity
        self.mode: flag indicating whether SVs are to be randomly generated or read in from VCF
        self.vcf_path: optional path that will be used if mode=="fixed"
        """
        if self.mode == "randomized":
            for sv_config in self.svs_config:
                if "avoid_intervals" in sv_config:
                    continue
                for num in range(sv_config["number"]):
                    # logic for placing events at intervals given in overlap bed file:
                    # for the first (sv_config["num_overlap"]) events, instantiate the SV at the next valid repeat elt interval
                    repeat_elt = None
                    elt_type = None
                    if self.overlap_events is not None:
                        sv_config_identifier = utils.get_sv_config_identifier(sv_config)
                        if sv_config_identifier in self.overlap_events.svtype_overlap_counts.keys():
                            repeat_elt, retrieved_type, elt_type = self.overlap_events.get_single_element_interval(
                                sv_config_identifier, sv_config, partial_overlap=False)
                        elif sv_config_identifier in self.overlap_events.svtype_partial_overlap_counts.keys():
                            repeat_elt, retrieved_type, elt_type = self.overlap_events.get_single_element_interval(
                                sv_config_identifier, sv_config, partial_overlap=True)
                        elif sv_config_identifier in self.overlap_events.svtype_alu_mediated_counts.keys():
                            repeat_elt, retrieved_type = self.overlap_events.get_alu_mediated_interval(sv_config_identifier)
                    if sv_config['type'] == Variant_Type.SNP:
                        sv = Structural_Variant(sv_type=sv_config["type"], mode=self.mode, length_ranges=[(1, 1)])
                    else:
                        sv = Structural_Variant(sv_type=sv_config["type"], mode=self.mode,
                                                length_ranges=sv_config["length_ranges"], source=sv_config["source"],
                                                target=sv_config["target"],
                                                overlap_event=(repeat_elt + (retrieved_type if elt_type in ['ALL', None] else elt_type,) if repeat_elt is not None else None),
                                                div_prob=(None if 'divergence_prob' not in sv_config.keys() else sv_config['divergence_prob']))

                    # For divergent repeat simulation, need div_dDUP to be homozygous
                    if self.sim_settings.get("homozygous_only", False) or random.randint(0, 1) or sv.type == Variant_Type.div_dDUP:
                        sv.ishomozygous = Zygosity.HOMOZYGOUS
                        sv.hap = [True, True]
                    else:
                        sv.ishomozygous = Zygosity.HETEROZYGOUS
                        sv.hap = random.choice([[True, False], [False, True]])

                    self.svs.append(sv)
            if not self.sim_settings["prioritize_top"]:
                random.shuffle(self.svs)
        else:  # <-- branch for mode == "fixed"
            self.process_vcf(self.vcf_path)

    def produce_variant_genome(self, fasta1_out, fasta2_out, ins_fasta, bedfile, stats_file=None, initial_reset=True,
                               verbose=False, export_to_file=True):
        """
        initial_reset: boolean to indicate if output file should be overwritten (True) or appended to (False)
        stats_file: whether a stats file summarizing SVs simulated will be generated in same directory the reference genome is located in
        """
        global time_start
        if initial_reset:
            utils.reset_file(fasta1_out)
            utils.reset_file(fasta2_out)
        ref_fasta = self.ref_fasta
        self.apply_transformations(ref_fasta)
        print("Finished SV placements and transformations in {} seconds".format(time.time() - time_start))
        time_start = time.time()
        active_svs = [sv for sv in self.svs if sv.active]
        print("Starting Export Process...")
        for x in range(2):
            edits_dict = dict()
            for id in self.order_ids:
                edits_dict[id] = []
            if x == 0:
                fasta_out = fasta1_out
            elif x == 1:
                fasta_out = fasta2_out
            for sv in active_svs:
                if sv.hap[x]:
                    for frag in sv.changed_fragments:
                        edits_dict[frag[0]].append(frag[1:])
            for id in edits_dict:
                edits_dict[id].sort()
                self.event_ranges[id].sort()
            self.log_to_file("Event Ranges: {}".format(self.event_ranges))
            self.log_to_file("Intervals for hap {}: {}".format(x, edits_dict))
            for id in self.order_ids:
                edits_x = edits_dict[id]
                utils.fail_if_any_overlapping(edits_x)
                self.formatter.export_variants_to_fasta(id, edits_x, fasta_out, ref_fasta, verbose=verbose)
                print("ID {} exported to fasta file {} in {} seconds".format(id, fasta_out, time.time() - time_start))
                time_start = time.time()
        if export_to_file:
            self.formatter.export_to_bedpe(active_svs, bedfile, ins_fasta=ins_fasta, reset_file=initial_reset)
            self.formatter.export_to_vcf(active_svs, self.stats, vcffile=bedfile[:-4]+'.vcf')
        if stats_file:
            self.stats.get_info(self.svs)
            self.stats.export_data(stats_file)

    def choose_rand_pos(self, svs, ref_fasta, verbose=False):
        """
        randomly positions SVs and stores reference fragments in SV events

        svs: list of Structural Variant objects
        ref_fasta: FastaFile with access to reference file
        """
        active_svs_total = 0
        inactive_svs_total = 0
        time_start_local = time.time()
        for sv in svs:
            tries = 0  # number of attempts to place sv randomly
            valid = False
            while not valid:
                tries += 1
                valid = True
                if tries > self.sim_settings["max_tries"]:
                    if self.sim_settings["fail_if_placement_issues"]:
                        raise Exception(
                            "Failed to simulate {}, {} / {} SVs successfully simulated (set fail_if_placement_issues "
                            "to False to override placement failures)".format(
                                sv, active_svs_total, len(svs)))
                    valid = False
                    break
                rand_id, chr_len, chr_event_ranges = self.get_rand_chr(check_size=sv.req_space,
                                                                       fixed_chrom=(None if sv.overlap_event is None
                                                                                    else sv.overlap_event[0]))
                if not (sv.dispersion_flip and sv.overlap_event is not None):
                    # if an overlap event is given, need to find the SV start position based on which fragment has been
                    # set to the overlap event interval
                    if sv.overlap_event is not None:
                        start_pos = 0
                        for frag in sv.source_events[::-1]:
                            if frag.start is not None:
                                start_pos = frag.start
                            else:
                                start_pos -= frag.length
                    else:
                        start_pos = random.randint(0, chr_len - sv.req_space)
                    # define the space in which SV operates
                    new_intervals = []  # tracks new ranges of blocks
                    sv.start, sv.start_chr = start_pos, rand_id
                    sv.end = sv.start + sv.req_space
                    block_start = sv.start
                else:
                    # to assign event "A" to a repeat interval in a flipped dispersion event, need to
                    # anchor the sv to the end of "A" and get the start position by subtracting off the total size
                    end_pos = int(sv.overlap_event[2])
                    start_pos = end_pos - sv.req_space
                    new_intervals = []
                    sv.start, sv.start_chr = start_pos, rand_id
                    sv.end = end_pos
                    block_start = sv.start

                for sv_event in sv.source_events:
                    sv_event.start, sv_event.end = start_pos, start_pos + sv_event.length
                    sv_event.source_chr = rand_id
                    frag = ref_fasta.fetch(rand_id, sv_event.start, sv_event.end)
                    sv_event.source_frag = frag
                    start_pos += sv_event.length

                    if sv_event.symbol.startswith(Symbols.DIS.value):
                        if utils.is_overlapping(chr_event_ranges, (block_start, sv_event.start)):
                            valid = False
                            break
                        new_intervals.append((block_start, sv_event.start))
                        block_start = sv_event.end
                    elif utils.percent_N(frag) > 0.05:
                        valid = False
                        break
                # catches the last (and perhaps only) block in sequence
                if utils.is_overlapping(chr_event_ranges, (block_start, sv.end)):
                    valid = False
                    continue
                else:
                    new_intervals.append((block_start, sv.end))

            # adds new SV to simulate only if chosen positions were valid
            if valid:
                active_svs_total += 1
                sv.active = True
                self.log_to_file("Intervals {} added to Chromosome \"{}\" for SV {}".format(new_intervals, rand_id, sv))
                chr_event_ranges.extend(new_intervals)
                # populates insertions with random sequence - these event symbols only show up in target transformation
                for event in sv.events_dict.values():
                    if event.source_frag is None and event.length > 0:
                        event.source_frag = utils.generate_seq(event.length)
                sv.assign_locations(sv.start)
            else:
                inactive_svs_total += 1
                if tries != self.sim_settings["max_tries"] + 1:
                    self.log_to_file("{} only got {} tries instead of the max {}".format(sv, tries, self.sim_settings[
                        "max_tries"] + 1), key="WARNING")

            time_dif = time.time() - time_start_local
            print(
                "{} / {} SVs successfully placed ========== {} / {} SVs unsuccessfully placed, {} tries, {} seconds".format(
                    active_svs_total, len(svs), inactive_svs_total, len(svs), tries, time_dif), end="\r")
            time_start_local = time.time()

    def apply_transformations(self, ref_fasta):
        """
        Randomly chooses positions for all SVs and carries out all edits
        Populates event classes within SVs with reference fragments and start & end positions
        Stores list of changes, which each have an interval and a sequence to substitute the reference frag with, in SV

        ref_fasta: FastaFile with access to reference
        mode: flag indicating whether we're adding SVs to the reference in a randomized or deterministic way
        """
        if self.mode == "randomized":
            # select random positions for SVs
            self.choose_rand_pos(self.svs, ref_fasta)
            print()

        total = 0
        for sv in self.svs:
            if sv.active:
                sv.change_fragment()
                total += 1
                self.log_to_file("Events Dict after all edits: {} ".format(sv.events_dict))

    def close(self):
        self.ref_fasta.close()

def run_insilicosv():
    sim_start = time.time()

    args = collect_args()
    if args["random_seed"]:
        random.seed(args["random_seed"])
        print(f"Random seed: {args['random_seed']}")
    fasta_in = args["ref"]
    yaml_in = args["config"]
    fasta1_out = args["hap1"]
    fasta2_out = args["hap2"]
    ins_fasta = args["ins_fasta"]
    bed_out = args["bedpe"]
    stats_file = args["stats"]
    log_file = args["log_file"]

    '''fasta_in = "debugging/inputs/test.fna"
    yaml_in = "debugging/inputs/par.yaml"
    fasta1_out = "debugging/inputs/test1_out.fna"
    fasta2_out = "debugging/inputs/test2_out.fna"
    bed_out = "debugging/inputs/out.bed"
    stats_file = "debugging/inputs/stats.txt"'''

    sim = SV_Simulator(args["ref"], args["config"], log_file=args["log_file"])
    sim.produce_variant_genome(args["hap1"], args["hap2"], args["ins_fasta"], args["bedpe"], args["stats"],
                               verbose=False)
    print("Simulation completed in {} seconds".format(time.time() - sim_start))

if __name__ == "__main__":
    run_insilicosv()



import os
from setuptools import setup

setup(
    name = "insilicosv",
    version = "0.0.4",
    description = ("Simulator of complex structural variants in the genome "),
    #license = "BSD",
    #keywords = "DNA simulation structural variant",
    url = "https://github.com/PopicLab/insilicoSV",

    packages=['insilicosv'],
    #long_description=read('README.md'),
    long_description='Simulator of complex genomic structural variants that support custom variant structures',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: BSD License",
    ],
)
